"""Function to search ArcticDEM and REMA index files.
"""

import os, warnings

from datetime import datetime
from typing import Optional, Literal, List

import pandas as pd
import geopandas as gpd

from pystac_client import Client
from shapely.geometry import box, polygon

SENSORS = ["WV03", "WV02", "WV01", "GE01"]
MONTHS = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]


def search(
    dataset: Optional[Literal["arcticdem", "rema"]] = None,
    dates: Optional[str | tuple] = None,
    months: Optional[int | tuple] = None,
    years: Optional[int | tuple] = None,
    baseline_max_hours: Optional[int] = None,
    bounds: Optional[tuple | polygon.Polygon] = None,
    min_aoi_frac: Optional[float] = None,
    min_aoi_km2: Optional[float] = None,
    sensors: Optional[str | tuple] = None,
    is_xtrack: Optional[bool] = None,
    is_lsf: Optional[bool] = None,
    rmse: Optional[float | tuple] = None,
    accuracy: Optional[float | tuple] = None,
    index_fpath: Optional[str] = None,
    include_custom_columns: Optional[bool] = True,
):
    """Efficiently search the ArcticDEM and REMA strip index files provided by the
    Polar Geospatial Center.

    :param dataset: Either 'arcticdem' or 'rema'. Case-insensitive. Must be provided
        if the dynamic STAC is being queried (default operation). If a local
        ``index_fpath`` is provided, then an attempt will be made to infer the
        dataset from the file name.
    :type dataset: str
    :param dates: Filter strips to a date range. Dates can be provided as a tuple of two
        strings, or a single string with a `/` seperator. Date strings must be
        interpetable by the pandas.to_datetime() tool. For an open search, one
        of the dates can be set to None.
    :type dates: str | tuple, optional
    :param months: Filter strips to only certain months. Provide as a tuple of integers
        (e.g. for June, July, August strips only, set ``months = [6,7,8]``).
    :type months: tuple, optional
    :param years: Filter strips to only certain years. Provide as a tuple of integers
        (e.g. for 2011 and 2021  only, set `years = [2011,2021]`).
    :type years: tuple, optional
    :param baseline_max_hours: Filter strips to only those constructed from stereopairs
        acquired less than the provided number of hours apart.
    :type baseline_max_hours: int, optional
    :param bounds: Filter to strips that intersect with bounds [xmin, ymin, xmax, ymax],
        in EPSG:3413 (ArcticDEM) or EPSG:3031 (REMA). Accepts a tuple or a
        ``shapely.Polygon`` geometry to extract bounds from.
    :type bounds: tuple | Polygon, optional
    :param min_aoi_frac: Filter to strips than cover more the defined fraction of the
        area of interest (defined by the `bounds` variable). Must be between 0 and 1,
        and ``bounds`` must be defined.
    :type min_aoi_frac: float, optional
    :param min_aoi_km2: Filter to strips than cover more the defined area of interest
        (defined by the `bounds` variable) in square kilometres.
    :type min_aoi_km2: float, optional
    :param sensors: Filter scenes to only those consrtructed from the provided
        satellite sensors. Full list is ["WV03", "WV02", "WV01", "GE01"]
    :type sensors: tuple, optional
    :param is_xtrack: Filter based on whether stereopairs are cross-track imagery.
        True = return only cross-track. False = return only non-cross-track.
    :type is_xtrack: bool, optional
    :param rmse: Filter based on the provided RMSE in metres from the PGC strip
        metadata. NB that this column can include NaN values (-1.0 or -9999.0) so
        the option is provided to include only a single value as an upper range
        (e.g. 2), or a tuple of two values in order to include a lower bound and
        filter NaN values (e.g [0, 2]).
    :type rmse: float | tuple, optional
    :param accuracy: Filter to strip accuracies based on the metadata-provided
        avg_expected_height_accuracy in metres. NB that this column can include
        NaN values (-1.0 or -9999.0) so the option is provided to include only a single
        value as an upper range (e.g. 2), or a tuple of two values in order to include
        a lower bound and filter NaN values (e.g [0, 2]).
    :type accuracy: float | tuple, optional
    :param index_fpath: Filepath to a local copy of the ArcticDEM or REMA strip index
        file, available from the Polar Geospatial Center. For speed, it is tecommended
        that the index is `.parquet` format. If none is provided, the function will
        instead interact with the PGC dynamic STAC API.
    :type index_fpath: str
    :param include_custom_columns: Whether to include custom columns generated by
        `pdemtools` in geopandas dataframe output. This includes "pdt_time1",
        "pdt_time2", "pdt_dem_baseline_hours", "pdt_time_mean", "pdt_year", "pdt_month",
        and "pdt_aoi_frac". Defaults to True
    :type include_custom_columns: bool, optional

    :returns: Strip index filtered to desired variables. If no strips are found,
        returns `None` with a warning.
    :rtype: geopandas.GeoDataFrame
    """

    # ----------------------------------------------------------------------- #
    # SANITISE AND PREPARE INPUTS
    # ----------------------------------------------------------------------- #

    # Sanitise inputs: dataset

    # If `dataset` not provided, perform a simple check for whether index is for
    # ArcticDEM or REMA by checking whether 'arcticdem' or 'rema' strings occur
    # in the  filename of the index file.
    if dataset is None:
        if index_fpath is not None:
            if "arcticdem" in os.path.basename(index_fpath).lower():
                dataset = "arcticdem"
            elif "rema" in os.path.basename(index_fpath).lower():
                dataset = "rema"
            else:
                raise ValueError(
                    "Cannot determine whether index file is ArcticDEM or REMA (please ensure 'arcticdem' or 'rema' is in filename, or provide `dataset` variable manually)"
                )
        else:
            raise ValueError("`dataset` must be either 'arcticdem' or 'rema'")

    # From dataset, set the collection and epsg
    if dataset.lower() == "arcticdem":
        collection = "arcticdem-strips-s2s041-2m"
        epsg = 3413

    elif dataset.lower() == "rema":
        collection = "rema-strips-s2s041-2m"
        epsg = 3031
    else:
        raise ValueError("`dataset` must be either 'arcticdem' or 'rema'")

    # Sanitise input: bounds
    if bounds is not None:
        # if type(bounds) is not polygon.Polygon:
        if not isinstance(bounds, polygon.Polygon):
            geom = box(*bounds)
        else:
            geom = bounds
        # convert geom to EPSG:4326
        geom_4326 = gpd.GeoDataFrame(geometry=[geom], crs=epsg).to_crs(4326).geometry[0]
    else:
        geom, geom_4326 = None, None

    # Sanitise input: min_aoi_frac
    if min_aoi_frac is not None:
        if bounds is None:
            raise ValueError("`bounds` variable must be provided to use `min_aoi_frac`")
        if (min_aoi_frac < 0) or (min_aoi_frac > 1):
            raise ValueError(
                f"`min_aoi_frac` must be a value between 0 and 1. Currently {min_aoi_frac}"
            )

    # Sanitise input: dates
    if dates is not None:
        if isinstance(dates, str):
            dates = dates.split("/")
            dates = [_correct_date_format(dates[0]), _correct_date_format(dates[1])]

        if len(dates) != 2:
            raise ValueError(
                "Date range must be tuple of two strings/None values, or a string of two dates/None values seperated by `/`"
            )

    # Sanitise input: years
    if years is not None:
        if not isinstance(years, list):
            if isinstance(years, int):
                years = [years]
            else:
                raise ValueError(
                    "Variables `years` must be integer year or list of integer years"
                )

    # Sanitise input: months
    if months is not None:
        if isinstance(months, int):
            months = [months]

        if any(m not in MONTHS for m in months):
            raise ValueError(f"`months` variables must be in range {MONTHS}")

    # Sanitise input: sensors
    if sensors is not None:
        if isinstance(sensors, str):
            sensors = [sensors]

        if any(s not in SENSORS for s in sensors):
            raise ValueError(f"`sensors` variables must be in list {SENSORS}")

    # Sanitise input: rmse
    if rmse is not None:
        if isinstance(rmse, int) or isinstance(rmse, float):
            rmse = [rmse]
        if len(rmse) > 2:
            raise ValueError("`accuracy` must be single value or tuple of length 2")

    # Sanitise input: accuracy
    if accuracy is not None:
        if isinstance(accuracy, int) or isinstance(accuracy, float):
            accuracy = [accuracy]
        if len(accuracy) > 2:
            raise ValueError("`accuracy` must be single value or tuple of length 2")

    # ----------------------------------------------------------------------- #
    # DOWNLOAD FROM STAC OR EXTRACT FROM INDEX FILE
    # ----------------------------------------------------------------------- #

    if index_fpath is None:

        if dataset is None:
            raise ValueError(
                "Must provide `dataset` to query PGC dynamic STAC - either 'arcticdem' or 'rema'"
            )

        gdf = _search_index_stac(
            collection,
            geom_4326,
            dates,
            years,
            is_xtrack,
            is_lsf,
            sensors,
        )

    else:

        gdf = _search_index_parquet(
            index_fpath,
            geom_4326,
            sensors,
            is_xtrack,
            is_lsf,
        )

    if gdf is None:
        warnings.warn(
            "No strips found matching search parameters. Returning `None`.",
            UserWarning,
            stacklevel=2,
        )
        return None

    # ----------------------------------------------------------------------- #
    # FILTER BASED ON DATETIMES
    # ----------------------------------------------------------------------- #

    # Construct additional pdt-specific columns
    gdf["dem_baseline_hours"] = (
        (gdf.datetime2 - gdf.datetime1)
        .values.astype("timedelta64[h]")
        .astype("float32")
    )
    gdf["dem_baseline_hours"] = abs(
        gdf["dem_baseline_hours"].values
    )  # Absolute d_t, in hours
    gdf["datetime_mean"] = gdf.datetime1 + (gdf.datetime2 - gdf.datetime1) / 2
    gdf["year"] = gdf.datetime_mean.dt.year
    gdf["month"] = gdf.datetime_mean.dt.month

    # Filter according to date (acqdate)
    if dates is not None:
        if dates[0] is not None:
            datetime1 = pd.to_datetime(dates[0])
            try:
                gdf = gdf[gdf["datetime1"] > datetime1]
            except:  # account for UTC-aware datetime storage in newer PGC index files.
                datetime1 = pd.to_datetime(dates[0], utc=True)
                gdf = gdf[gdf["datetime1"] > datetime1]

        if dates[1] is not None:
            datetime2 = pd.to_datetime(dates[1])
            try:
                gdf = gdf[gdf["datetime2"] < datetime2]
            except:  # account for UTC-aware datetime storage in newer PGC index files.
                datetime2 = pd.to_datetime(dates[1], utc=True)
                gdf = gdf[gdf["datetime2"] < datetime2]
            # can probably make this more efficient than just try-except - e.g. check if
            # date column is utc-aware and if so, make datetime utc-aware too.

    # Filter to months
    if months is not None:
        gdf = gdf[(gdf["month"].isin(months))]

    # Filter to years
    if years is not None:
        gdf = gdf[(gdf["year"].isin(years))]

    # Filter to time seperations
    if baseline_max_hours is not None:
        gdf = gdf[(gdf["dem_baseline_hours"] <= baseline_max_hours)]

    # Filter to RMSE range
    if rmse is not None:
        if len(rmse) == 1:
            gdf = gdf[(gdf["rmse"] <= rmse[0])]
        else:
            gdf = gdf[(gdf["rmse"] >= rmse[0]) & (gdf["rmse"] <= rmse[1])]

    # Filter to accuracy range
    if accuracy is not None:
        if len(accuracy) == 1:
            gdf = gdf[(gdf["rmse"] <= accuracy[0])]
        else:
            gdf = gdf[(gdf["rmse"] >= accuracy[0]) & (gdf["rmse"] <= accuracy[1])]

    if len(gdf) == 0:
        warnings.warn(
            "No strips found matching search parameters. Returning `None`.",
            UserWarning,
            stacklevel=2,
        )
        return None

    # ----------------------------------------------------------------------- #
    # AOI FRACTIONS
    # ----------------------------------------------------------------------- #

    # Convert from EPSG:4326 to appropriate ArcticDEM/REMA crs:
    gdf = gdf.to_crs(epsg)

    # Clip geometries to bounds of AOI
    if bounds is not None:
        gdf = gpd.clip(gdf, geom)

    # Filter based on intersection area
    if min_aoi_km2 is not None:
        gdf["aoi_km2"] = gdf.area / 1e6
        gdf = gdf[gdf["aoi_km2"] > min_aoi_km2]

    # Filter based on intersection fraction
    if min_aoi_frac is not None:
        aoi_area = geom.area
        gdf["aoi_frac"] = gdf.area / aoi_area
        gdf = gdf[gdf["aoi_frac"] > min_aoi_frac]

    if len(gdf) == 0:
        warnings.warn(
            "No strips found matching search parameters. Returning `None`.",
            UserWarning,
            stacklevel=2,
        )
        return None

    # ----------------------------------------------------------------------- #
    # WRAP UP AND EXPORT
    # ----------------------------------------------------------------------- #

    # Sort values by date as default
    gdf.sort_values("datetime1", inplace=True)
    gdf.reset_index(drop=True, inplace=True)

    # If include_custom_columns is True, rename columns to begin with "pdt_"
    if include_custom_columns:
        gdf = gdf.rename(
            columns={
                "id": "pdt_id",
                "datetime1": "pdt_datetime1",
                "datetime2": "pdt_datetime2",
                "sat1": "pdt_sat1",
                "sat2": "pdt_sat2",
                "dem_baseline_hours": "pdt_dem_baseline_hours",
                "datetime_mean": "pdt_datetime_mean",
                "year": "pdt_year",
                "month": "pdt_month",
                "aoi_frac": "pdt_aoi_frac",
            }
        )

    # Else, drop the created columns
    else:
        gdf = gdf.drop(
            [
                "datetime1",
                "datetime2",
                "sat1",
                "sat2",
                "dem_baseline_hrs",
                "time_mean",
                "month",
                "year",
                "aoi_frac",
            ],
            axis=1,
            errors="ignore",
        )

    return gdf


def _search_index_stac(
    collection: Literal["arcticdem-strips-s2s041-2m", "rema-strips-s2s041-2m"],
    geom_4326: polygon.Polygon,
    dates: Optional[List[str]] = None,
    years: Optional[List[int]] = None,
    is_xtrack: Optional[bool] = None,
    is_lsf: Optional[bool] = None,
    sensors: Optional[List[str]] = None,
):

    # Construct datetime query:
    if dates is not None:
        date_search = dates
    elif years is not None:
        date_search = [f"{min(years)}", f"{max(years)}"]
    else:
        date_search = None

    # Construct the query
    query = {}

    if is_xtrack is not None:
        query["pgc:is_xtrack"] = {"eq": is_xtrack}

    if is_lsf is not None:
        query["pgc:is_lsf"] = {"eq": is_lsf}

    # Search the stac
    client = Client.open("https://stac.pgc.umn.edu/api/v1/")

    search_results = client.search(
        collections=collection,
        intersects=geom_4326,
        datetime=date_search,
        query=query,
    )

    # Transform into geopandas
    items = search_results.item_collection()

    # Close the client
    del client

    if len(items) == 0:
        return None

    # Construct geodataframe
    gdf = gpd.GeoDataFrame.from_features(
        items.to_dict(),
        crs="epsg:4326",
    )

    # Remove 'pgc:' column prefix and remove analytically useless columns
    gdf.columns = gdf.columns.str.replace("^pgc:", "", regex=True)
    gdf = gdf.drop(["description"], axis=1, errors="ignore")

    # Construct appropriate hrefs, and drop reference column
    gdf["href_json"] = [i.self_href + ".json" for i in items]
    gdf["href_dem"] = [i.assets["dem"].href for i in items]
    gdf["href_mask"] = [i.assets["mask"].href for i in items]
    gdf["href_readme"] = [i.assets["readme"].href for i in items]
    gdf["href_matchtag"] = [i.assets["matchtag"].href for i in items]
    gdf["href_metadata"] = [i.assets["metadata"].href for i in items]
    gdf["href_hillshade"] = [i.assets["hillshade"].href for i in items]
    gdf["href_hillshade_masked"] = [i.assets["hillshade_masked"].href for i in items]

    # Filter based on provided instruments
    if sensors is not None:
        gdf = gdf[
            gdf["instruments"].apply(
                lambda instruments: all(inst in sensors for inst in instruments)
            )
        ]

    # Extract dates to standardised column name for main function
    gdf["id"] = gdf.title
    gdf["datetime1"] = pd.to_datetime(gdf.start_datetime)
    gdf["datetime2"] = pd.to_datetime(gdf.end_datetime)
    gdf[["sat1", "sat2"]] = gdf["instruments"].apply(pd.Series)

    return gdf


def _search_index_parquet(
    index_fpath: str,
    geom_4326: Optional[polygon.Polygon] = None,
    sensors: Optional[List[str]] = None,
    is_xtrack: Optional[bool] = None,
    is_lsf: Optional[bool] = None,
):

    # Open the index geometry file, according to file type
    _, extension = os.path.splitext(index_fpath)

    if extension == ".parquet":
        # print("reading")
        gdf = gpd.read_parquet(index_fpath)
        if geom_4326 is not None:
            gdf = gdf[gdf.intersects(geom_4326)]

    else:
        print(
            "For quicker searching, it is highly recommended to store strip index files as a `.parquet` or `.feather` format."
        )
        if geom_4326 is None:
            gdf = gpd.read_file(index_fpath)
        else:
            gdf = gpd.read_file(index_fpath, intersects=geom_4326)

    # Filter to only selected sensors
    if sensors is not None:
        gdf = gdf[(gdf["sensor1"].isin(sensors)) & (gdf["sensor2"].isin(sensors))]

    # Filter to crosstrack
    if is_xtrack is not None:
        gdf = gdf[(gdf["is_xtrack"] == is_xtrack)]

    if is_lsf is not None:
        gdf = gdf[(gdf["is_lsf"] == is_lsf)]

    if len(gdf) == 0:
        return None

    # Extract hrefs from search
    gdf["href_json"] = "https://" + gdf["s3url"].str.split("/external/").str[1]
    gdf["href_dem"] = gdf["href_json"].str.replace(".json", "_dem.tif", regex=False)
    gdf["href_mask"] = gdf["href_json"].str.replace(
        ".json", "_bitmask.tif", regex=False
    )
    gdf["href_readme"] = gdf["href_json"].str.replace(
        ".json", "_readme.txt", regex=False
    )
    gdf["href_matchtag"] = gdf["href_json"].str.replace(
        ".json", "_matchtag.tif", regex=False
    )
    gdf["href_metadata"] = gdf["href_json"].str.replace(
        ".json", "_mdf.txt", regex=False
    )
    gdf["href_hillshade"] = gdf["href_json"].str.replace(
        ".json", "_dem_10m_shade.tif", regex=False
    )
    gdf["href_hillshade_masked"] = gdf["href_json"].str.replace(
        ".json", "_dem_10m_shade_masked.tif", regex=False
    )

    # Extract dates to standardised column name for main function
    gdf["id"] = gdf.dem_id
    gdf["datetime1"] = pd.to_datetime(gdf.acqdate1)
    gdf["datetime2"] = pd.to_datetime(gdf.acqdate2)
    gdf["sat1"] = gdf.sensor1
    gdf["sat2"] = gdf.sensor2

    return gdf


def _correct_date_format(date_str):
    # catch str date format of YYYYMMDD and convert to YYYY-MM-DD if so.
    try:
        date = datetime.strptime(date_str, "%Y%m%d")
        return date.strftime("%Y-%m-%d")
    except ValueError:
        return date_str



"""
This module contains the functions necessary for coregistration according to Nuth and
Kääb (2011), based on the coregisterdems() function by Erik Husby at line 720 of the 
PGC setsm_processing scenes2strips.py module.

Original available at:
https://github.com/PolarGeospatialCenter/setsm_postprocessing_python/blob/fd36fd54933ec43f587902a4fdcd1acbd90951c2/lib/scenes2strips.py
"""

import operator
from typing import Optional
from warnings import warn

import numpy as np
import geopandas as gpd

from scipy import stats
from scipy.interpolate import interpn
from osgeo import gdal, gdal_array
from xarray import Dataset

from ._geomorphometry import p_f, q_f
from ._utils import get_resolution

# suppress printing out warnings
np.seterr(divide="ignore", invalid="ignore")
gdal.UseExceptions()


def coregister(
    dem: np.ndarray,
    reference: np.ndarray | list[np.ndarray],
    x: np.ndarray,
    y: np.ndarray,
    res: float,
    mask: Optional[np.ndarray] = None,
    max_horiz_offset: Optional[float] = 50,
    rmse_step_thresh: Optional[float] = -0.001,
    max_iterations: Optional[int] = 5,
    verbose: Optional[bool] = True,
):
    """Backend coregistration routine, coregistering DEMs against either grid
    or point data according to Nuth and Kääb (2011). Accepts as a reference
    either another DEM (of the same size as the target DEM) or point data (for
    e.g. ICESat-2 altimetry) as a three-item list of [x, y, z] data.
    """
    # verbose print lamdba function
    print_verbose = lambda msg: print(msg) if verbose else None

    dem2 = dem  # .copy()

    # Filter points to only valid and unmasked data
    if isinstance(reference, (list, tuple)):

        coreg_type = "points"
        # grid_x, grid_y = np.meshgrid(x, y, indexing='ij')

        points_x = reference[0]
        points_y = reference[1]
        points_h = reference[2]
        points_xy = list(zip(points_y, points_x))

        # Mask invalid (and masked) values here
        points_dem = interpn(
            points=(y, x),
            values=dem2,
            xi=points_xy,
            method="nearest",
            bounds_error=False,
        )
        points_valid = ~np.isnan(points_dem)

        if mask is None:
            points_h = points_h[points_valid == 1]
            points_x = points_x[points_valid == 1]
            points_y = points_y[points_valid == 1]

        if mask is not None:
            points_mask = interpn(
                points=(y, x),
                values=mask,
                xi=points_xy,
                method="nearest",
                bounds_error=False,
            )
            points_h = points_h[(points_valid == 1) & (points_mask == 1)]
            points_x = points_x[(points_valid == 1) & (points_mask == 1)]
            points_y = points_y[(points_valid == 1) & (points_mask == 1)]

        points_xy = list(zip(points_y, points_x))
        points_n = len(points_h)

    else:
        coreg_type = "dem"

    # initial trans and RMSE settings
    p = np.zeros((3, 1))  # p  is prior iteration trans var
    pn = p.copy()  # pn is current iteration trans var
    perr = np.zeros((3, 1))  # perr is prior iteration regression errors
    pnerr = perr.copy()  # pnerr is current iteration regression errors
    d0 = np.inf  # initial RMSE

    # Edge case markers
    meddz = None
    return_meddz = False
    critical_failure = False

    it = 0
    while True:
        it += 1
        print_verbose(f"Planimetric Correction Iteration {it}")

        print_verbose(f"Offset (z,x,y): {pn[0, 0]:.3f}, {pn[1, 0]:.3f}, {pn[2, 0]:.3f}")
        # print(f"pn: {pn}")

        # Break loop if conditions reached
        if np.any(np.abs(pn[1:]) > max_horiz_offset):
            print(
                f"Maximum horizontal offset ({max_horiz_offset}) exceeded. "
                "Consider raising the threshold if offsets are large."
            )
            return_meddz = True
            break

        # Apply offsets
        if pn[1] != 0 and pn[2] != 0:
            dem2n = shift_dem(dem2, pn.T[0], x, y, verbose=verbose).astype("float32")
        else:
            dem2n = dem2 - pn[0].astype("float32")

        # # Calculate slopes - original method from PGC
        # sy, sx = np.gradient(dem2n, res)
        # sx = -sx

        # Calculate slope - using Florinsky slope method (p = sx, q = sy)
        sy = q_f(dem2n, res)

        sx = p_f(dem2n, res)
        sy = -sy
        sx = -sx

        if coreg_type == "dem":
            # Difference grids.
            dz = dem2n - reference
            # Mask (in full script, both m1 and m2 are applied)
            dz[mask == 0] = np.nan

        elif coreg_type == "points":
            # Get relevant dz, sy, sx values at the icesat-2 coordinates
            # xda = zero_xda + dem2n
            # dem2n_coords = xda.interp(coords_ds, method='linear').values
            dem2n_coords = interpn(
                points=(y, x), values=dem2n, xi=points_xy, method="linear"
            )
            dz = dem2n_coords - points_h

            # xda = zero_xda + sy
            # sy = xda.interp(coords_ds, method='linear').values
            sy = interpn(points=(y, x), values=sy, xi=points_xy, method="linear")

            # xda = zero_xda + sx
            # sx = xda.interp(coords_ds, method='linear').values
            sx = interpn(points=(y, x), values=sx, xi=points_xy, method="linear")

        else:
            raise ValueError(
                "coreg_type must be 'dem' or 'points'. " f"Received {coreg_type}."
            )

        # If no overlap between scenes, break the loop
        if np.all(np.isnan(dz)):
            print("No overlapping data between reference and target datasets.")
            critical_failure = True
            break

        # Filter NaNs and outliers.
        n = (
            ~np.isnan(sx)
            & ~np.isnan(sy)
            & (np.abs(dz - np.nanmedian(dz)) <= 3 * np.nanstd(dz))
        )
        n_count = np.count_nonzero(n)

        if n_count < 10:
            print(f"Too few ({n_count}) registration points: 10 required")
            critical_failure = True
            break

        # Get RMSE
        d1 = np.sqrt(np.mean(np.power(dz[n], 2)))
        print_verbose(f"RMSE = {d1}")

        # Keep median dz if first iteration.
        if it == 1:
            meddz = np.median(dz[n])
            meddz_err = np.std(dz[n] / np.sqrt(n_count))
            d00 = np.sqrt(np.mean(np.power(dz[n] - meddz, 2)))

        # Get improvement in RMSE
        rmse_step = d1 - d0  # initial d0 == inf

        # break if rmse above threshold
        if rmse_step > rmse_step_thresh or np.isnan(d0):
            print_verbose(
                f"RMSE step in this iteration ({rmse_step:.5f}) is above threshold "
                f"({rmse_step_thresh}), stopping and returning values of prior iteration."
            )
            # If fails after first registration attempt,
            # set dx and dy to zero and subtract the median offset.
            if it == 2:
                print("Second iteration regression failure")
                return_meddz = True
            break
        elif it == max_iterations:
            print_verbose(f"Maximum number of iterations ({max_iterations}) reached")
            break

        # Keep this adjustment.
        dem2out = dem2n.copy()
        p = pn.copy()
        perr = pnerr.copy()
        d0 = d1

        # Build design matrix.
        X = np.column_stack((np.ones(n_count, dtype=np.float32), sx[n], sy[n]))
        sx, sy = None, None  # release for data amangement

        # Solve for new adjustment.
        p1 = np.reshape(np.linalg.lstsq(X, dz[n], rcond=None)[0], (-1, 1))

        # Calculate p errors.
        _, R = np.linalg.qr(X)
        RI = np.linalg.lstsq(R, np.identity(3, dtype=np.float32), rcond=None)[0]
        nu = X.shape[0] - X.shape[1]  # residual degrees of freedom
        yhat = np.matmul(X, p1)  # predicted responses at each data point
        r = dz[n] - yhat.T[0]  # residuals
        normr = np.linalg.norm(r)

        dz = None  # release for memory managment

        rmse = normr / np.sqrt(nu)
        tval = stats.t.ppf((1 - 0.32 / 2), nu)

        se = rmse * np.sqrt(np.sum(np.square(np.abs(RI)), axis=1, keepdims=True))
        p1err = tval * se

        # Update shifts.
        pn = p + p1
        pnerr = np.sqrt(np.square(perr) + np.square(p1err))

        # END OF LOOP

    if return_meddz:
        print(f"Returning median vertical offset: {meddz:.3f}")
        dem2out = dem2 - meddz
        p = np.array([[meddz, 0, 0]]).T
        perr = np.array([[meddz_err, 0, 0]]).T
        d0 = d00
        status = "dz_only"

    elif critical_failure:
        print("Regression critical failure, returning original DEM with no translation")
        dem2out = dem2
        p = np.full((3, 1), np.nan)
        perr = np.full((3, 1), np.nan)
        d0 = np.nan
        status = "failed"
        points_n = None

    else:
        status = "coregistered"

    print(f"Final offset (z,x,y): {p[0, 0]:.3f}, {p[1, 0]:.3f}, {p[2, 0]:.3f}")
    print(f"Final RMSE = {d0}")

    # Construct metadata:
    metadata_dict = {
        "coreg_status": status,
        "x_offset": p[1, 0],
        "y_offset": p[2, 0],
        "z_offset": p[0, 0],
        "x_offset_err": perr[1, 0],
        "y_offset_err": perr[2, 0],
        "z_offset_err": perr[0, 0],
        "rmse": d0,
    }

    if coreg_type == "points":
        metadata_dict["points_n"] = points_n

    # Convert all numerical values to regular Python floats
    metadata_dict = {
        key: float(value) if isinstance(value, (np.float64, np.float32)) else value
        for key, value in metadata_dict.items()
    }

    return dem2out, metadata_dict


def shift_dem(dem, trans, x, y, verbose=True):
    """
    Shifts DEM according to translation factors ascertained in coregisterdems function

    INPUT
    dem: DEM to correct (dem2 in coregisterdems parlance)
    trans: (dz, dx, dy) translation parameters

    """

    # print("Translating in Z direction")
    dem_coreg = dem - trans[0]

    if (trans[1] == 0) & (trans[2] == 0):
        pass
        # print("XY translation variables == 0. Skipping horizontal translation.")
    else:
        # print("Translating in XY direction.")
        if verbose:
            print(f"Translating: {trans[0]:.2f} Z, {trans[1]:.2f} X, {trans[2]:.2f} Y")

        # Interpolation grid
        #     x, y = np.meshgrid(np.arange(dem.shape[1]), np.arange(dem.shape[0]))
        # x = np.arange(dem.shape[1])
        # y = np.arange(dem.shape[0])
        xi = x - trans[1]
        yi = y - trans[2]
        # xi = x + trans[2]
        # yi = y + trans[1]

        # Check that uniform spacing is maintained (sometimes rounding errors).
        if len(np.unique(np.diff(xi))) > 1:
            xi = np.round(xi, 4)
        if len(np.unique(np.diff(yi))) > 1:
            yi = np.round(yi, 4)

        # Correct in xy dimensions - linearly interpolating floating data to reference grid
        dem_coreg = interp2_gdal(
            xi, yi, dem_coreg, x, y, "linear", extrapolate=False, oob_val=np.nan
        )

    return dem_coreg


def interp2_gdal(X, Y, Z, Xi, Yi, interp_str, extrapolate=False, oob_val=np.nan):
    """
    Resample array data from one set of x-y grid coordinates to another.
    From PGC's raster_array_tools.py

    INPUT:
    X, Y: Original grid coordinates (meshgrid)
    Z: Array to be redampled.
    Zi, Yi: New grid coordinated (meshgrid)
    Extrapolate: Interpolate values for pixels that fall outside range of old grid coords. If False, set values to `oob_val`
    oob_val: Fill regions of output array where new grid coords fall outside range of old grid coords.

    OUTPUT:
    Zi: The resampled array.
    """

    dtype_gdal, promote_dtype = dtype_np2gdal(Z.dtype)
    if promote_dtype is not None:
        Z = Z.astype(promote_dtype)

    interp_gdal = interp_str2gdal(interp_str)

    mem_drv = gdal.GetDriverByName("MEM")

    ds_in = mem_drv.Create("", X.size, Y.size, 1, dtype_gdal)
    ds_in.SetGeoTransform((X[0], X[1] - X[0], 0, Y[0], 0, Y[1] - Y[0]))
    ds_in.GetRasterBand(1).WriteArray(Z)

    ds_out = mem_drv.Create("", Xi.size, Yi.size, 1, dtype_gdal)
    ds_out.SetGeoTransform((Xi[0], Xi[1] - Xi[0], 0, Yi[0], 0, Yi[1] - Yi[0]))

    gdal.ReprojectImage(
        ds_in,
        ds_out,
        "",
        "",
        interp_gdal,
    )

    Zi = ds_out.GetRasterBand(1).ReadAsArray()

    if not extrapolate:
        interp2_fill_oob(X, Y, Zi, Xi, Yi, oob_val)

    ds_in, ds_out = None, None

    return Zi


def dtype_np2gdal(dtype_np):
    """
    Convert NumPy data type to GDAL data type.
    """

    if dtype_np == bool:  # np.bool:
        promote_dtype = np.uint8
    elif dtype_np == np.int8:
        promote_dtype = np.int16
    elif dtype_np == np.float16:
        promote_dtype = np.float32
    else:
        promote_dtype = None

    if promote_dtype is not None:
        warn(
            f"NumPy array data type ({dtype_np}) does not have equivalent GDAL "
            "data type and is not supported, but can be safely promoted to "
            f"{promote_dtype(1).dtype}",
            UserWarning,
            stacklevel=2,
        )
        dtype_np = promote_dtype

    dtype_gdal = gdal_array.NumericTypeCodeToGDALTypeCode(dtype_np)
    if dtype_gdal is None:
        raise InvalidArgumentError(
            f"NumPy array data type ({dtype_np}) does not have equivalent "
            "GDAL data type and is not supported",
            stacklevel=2,
        )

    return dtype_gdal, promote_dtype


def interp_str2gdal(interp_str):
    """
    Turrns `interp_str` string into a GDAL object for use in gdal.ReprojectImage.
    Accepts "nearest", "linear", "cubic", "spline", "lanczos", "average", "mode".
    """

    interp_choices = (
        "nearest",
        "linear",
        "cubic",
        "spline",
        "lanczos",
        "average",
        "mode",
    )

    interp_dict = {
        "nearest": gdal.GRA_NearestNeighbour,
        "linear": gdal.GRA_Bilinear,
        "bilinear": gdal.GRA_Bilinear,
        "cubic": gdal.GRA_Cubic,
        "bicubic": gdal.GRA_Cubic,
        "spline": gdal.GRA_CubicSpline,
        "lanczos": gdal.GRA_Lanczos,
        "average": gdal.GRA_Average,
        "mode": gdal.GRA_Mode,
    }

    if interp_str not in interp_dict:
        raise ValueError(
            f"`interp` must be one of {interp_choices}, but was '{interp_str}'"
        )

    return interp_dict[interp_str]


def interp2_fill_oob(X, Y, Zi, Xi, Yi, fillval=np.nan, coord_grace=True):
    # Rows and columns of Zi outside the domain of Z are made NaN.
    # Assume X and Y coordinates are monotonically increasing/decreasing
    # so hopefully we only need to work a short way inwards from the edges.

    Xi_size = Xi.size
    Yi_size = Yi.size

    Xmin = min(X[0], X[-1])
    Ymax = max(Y[0], Y[-1])

    x_lfttest_val = X[0]
    x_rgttest_val = X[-1]
    y_toptest_val = Y[0]
    y_bottest_val = Y[-1]

    if x_lfttest_val == Xmin:
        # X-coords increase from left to right.
        x_lfttest_op = operator.lt
        x_rgttest_op = operator.gt
    else:
        # X-coords decrease from left to right.
        x_lfttest_op = operator.gt
        x_rgttest_op = operator.lt

    if y_toptest_val == Ymax:
        # Y-coords decrease from top to bottom.
        y_toptest_op = operator.gt
        y_bottest_op = operator.lt
    else:
        # Y-coords increase from top to bottom.
        y_toptest_op = operator.lt
        y_bottest_op = operator.gt

    if coord_grace:
        x_grace = (X[1] - X[0]) / 64
        y_grace = (Y[1] - Y[0]) / 16
        x_lfttest_val -= x_grace
        x_rgttest_val += x_grace
        y_toptest_val -= y_grace
        y_bottest_val += y_grace
        x_lfttest_op = operator.le if x_lfttest_op(0, 1) else operator.ge
        x_rgttest_op = operator.le if x_rgttest_op(0, 1) else operator.ge
        y_toptest_op = operator.le if y_toptest_op(0, 1) else operator.ge
        y_bottest_op = operator.le if y_bottest_op(0, 1) else operator.ge

    i = 0
    while x_lfttest_op(Xi[i], x_lfttest_val) and i < Xi_size:
        Zi[:, i] = fillval
        i += 1
    i = -1
    while x_rgttest_op(Xi[i], x_rgttest_val) and i >= -Xi_size:
        Zi[:, i] = fillval
        i -= 1
    j = 0
    while y_toptest_op(Yi[j], y_toptest_val) and j < Yi_size:
        Zi[j, :] = fillval
        j += 1
    j = -1
    while y_bottest_op(Yi[j], y_bottest_val) and j >= -Yi_size:
        Zi[j, :] = fillval
        j -= 1

    return Zi



import pytest

import pdemtools as pdt

from numpy.testing import assert_allclose


def test_download():

    # Test download a small section of Store Glacier
    bounds = (-205650, -2132203, -204722, -2131404)
    expected_mean_elev = 263.5019226074219

    dem = pdt.load.mosaic(
        dataset="arcticdem", bounds=bounds, resolution=32, version="v4.1"
    )

    mean_elev = dem.mean().item()

    assert_allclose(mean_elev, expected_mean_elev, rtol=1e-5)


if __name__ == "__main__":
    test_download()



"""This module contains functions necessary extracting relevant datasets for DEM
processing (geoids, masks, etc), resampled to match the DEM xarray object.
"""

import datetime
from typing import Optional, Literal
from warnings import warn

import rioxarray as rxr
import geopandas as gpd
import sliderule

from shapely.geometry import box

from rasterio.enums import Resampling
from xarray import DataArray, Dataset
from geopandas.geodataframe import GeoDataFrame
from sliderule import sliderule, icesat2
from pandas import to_datetime

from ._utils import clip, get_resolution

# import os
# from shapely.geometry.polygon import Polygon


def geoid_from_bedmachine(bm_fpath: str, target_rxd: DataArray) -> DataArray:
    """Extracts the BedMachine geoid (EIGEN-6C4), bilinearly resampled to match the
    target dataset.

    :param bm_fpath: Filepath to BedMachine dataset, defaults to None
    :type bm_fpath: str
    :param target_rxd: (rio)xarray dataset that BedMachine will be resampled to match
    :type target_rxd: DataArray

    :returns: geoid for the target_rxd region as an xarray DataArray
    :rtype: DataArray"""

    geoid = rxr.open_rasterio(f"{bm_fpath}")["geoid"]
    geoid_crs = geoid.rio.crs
    geoid = geoid.squeeze().astype("float32").rio.write_crs(geoid_crs)
    geoid = geoid.rio.reproject_match(target_rxd, Resampling.bilinear)

    return geoid.squeeze()


def geoid_from_raster(fpath: str, target_rxd: DataArray = None) -> DataArray:
    """Extracts an arbritary geoid stored as a raster dataset, bilinearly resampled to
    match the target dataset.

    :param fpath: Filepath to geoid raster
    :type fpath: str
    :param target_rxd: (rio)xarray dataset/array that the raster will be resampled to
        match. Optional, defaults to None
    :type target_rxd: DataArray

    :returns: geoid for the target_rxd region as an xarray DataArray
    :rtype: DataArray
    """

    geoid = rxr.open_rasterio(f"{fpath}")
    geoid_crs = geoid.rio.crs
    geoid = geoid.squeeze().astype("float32").rio.write_crs(geoid_crs)

    if target_rxd != None:
        geoid = geoid.rio.reproject_match(target_rxd, Resampling.bilinear)

    return geoid.squeeze()


def bedrock_mask_from_vector(
    vector: str | GeoDataFrame, target_rxd: DataArray
) -> DataArray:
    """Construct boolean bedrock mask from a Geopandas vector file of bedrock areas and
    a given target rioxarray dataset. Returns mask where bedrock values are 1 and
    outside are 0.

    :param vector: either a GeoPandas GeoDataFrame of bedrock areas, or a filepath to
        a Geopandas-readable vector file (e.g. shapefile, geopackage, etc.).
    :type vector_fpath: str | GeoDataFrame
    :param target_rxd: (rio)xarray dataset that BedMachine will be resampled to match
    :type target_rxd: DataArray

    :returns: bedrock mask for the target_rxd region as a (rio)xarray DataArray
    :rtype: DataArray
    """

    if type(vector) == str:
        gdf_clip = gpd.read_file(vector)
    elif type(vector) == GeoDataFrame:
        gdf_clip = vector
    else:
        raise ValueError(
            "Input `vector` must be either a filepath string or GeoPandas GeoDataFrame"
        )
    target_rxd = target_rxd.rio.write_nodata(-9999)  # Enforce -9999 as nodata value
    target_clip = target_rxd.rio.clip(gdf_clip.geometry.values, drop=False)
    return (target_clip.where(target_clip != -9999) * 0 + 1).fillna(0).squeeze()


def bedrock_mask_from_bedmachine(bm_fpath: str, target_rxd: DataArray) -> DataArray:
    """Construct boolean bedrock mask from bedmachine and a given target rioxarray
    dataset. Returns mask where bedrock values are 1 and outside are 0.

    :param bm_fpath: Filepath to BedMachine dataset, defaults to None
    :type bm_fpath: str
    :param target_rxd: (rio)xarray dataset that BedMachine will be resampled to match
    :type target_rxd: DataArray

    :returns: bedrock mask for the target_rxd region as a (rio)xarray DataArray
    :rtype: DataArray
    """

    # Open geoid
    mask = rxr.open_rasterio(f"{bm_fpath}")["mask"]
    mask_crs = mask.rio.crs

    # Get geoid-projected geometry of the extent of the target dataset,
    # with a bit of a buffer for safety
    clip_bounds = (
        gpd.GeoDataFrame(
            geometry=[box(*target_rxd.rio.bounds())], crs=target_rxd.rio.crs
        )
        .to_crs(mask_crs)
        .buffer(1000)
        .total_bounds
    )

    # Clip geoid to smaller extent
    mask = clip(mask, clip_bounds)

    # bedmachine mask sets ice-free land to 1. Set everything else to zero.
    mask = mask.where(mask == 1, other=0)

    # Reample using bilinear resampling rather than nearest neighbour, as mask
    # resolution is much larger than DEM resolution
    mask = mask.squeeze().astype("float32").rio.write_crs(mask_crs)
    mask = mask.rio.reproject_match(
        match_data_array=target_rxd, resampling=Resampling.bilinear
    )

    # Set interpolated values to boolean (1/0) values again
    mask = mask.round()

    return mask.squeeze()


def icesat2_atl06(
    target_rxd: DataArray | tuple,
    date: str | datetime.date,
    stable_mask: Optional[DataArray] = None,
    days_r: Optional[int] = [7, 14, 28, 45],
    min_is2_points: Optional[int] = 100,
    filter_points: Optional[bool] = True,
    include_dt_column: Optional[bool] = True,
    epsg: Optional[int] = None,
    sr_srt: Optional[int] = 3,
    sr_cnf: Optional[int] = 1,
    sr_ats: Optional[float] = 20.0,
    sr_cnt: Optional[int] = 10,
    sr_len: Optional[float] = 20.0,
    sr_res: Optional[float] = 10.0,
    sr_track: Optional[int] = 0,
    sr_sigma_r_max: Optional[int] = 5,
):
    """Get ICESat-2 ATL06 data for a given target rioxarray dataset. Returns
    geopandas dataframe of ATL06-SR, processed and downloaded using the
    `sliderule` package.

    Parameters beginning with `sr_` are the sliderule parameters for
    `sliderule.icesat2.atl06p()`. All parameters are left to default, apart from
    the length `len` (20 m rather than 40 m) and step distance `res` (10 m rather
    than 20 m), in order to increase the resolution of the returned data to
    approximate the 10 m resolution of the default coregsitration DEM resolution.
    More information on parameters  are provided here:
    https://slideruleearth.io/web/rtd/user_guide/ICESat-2.html

    :param target_rxd: (rio)xarray dataset that ICESat-2 will be downloaded for.
        Also accepts a tuple of bounds in format [xmin, ymin, xmax, ymax], if the
        `epsg` parameter is provided.
    :type target_rxd: DataArray
    :param date: Date of dataset, as an interpretable string or datetime.date object.
    :type date: str | datetime.date
    :param stable_mask: (rio)xarray mask dataset where 1=valid, 0=masked. Will be
        used to spatially filter the ICESat-2 data. Defaults to None.
    :type stable_mask: DataArray
    :param days_r: Day range within which to search for ICESat-2 data. Should be
        ascending list. Defaults to [7, 14, 28, 45], meaning that all ICESat-2 data
        within 7 days will be returned, and then, if not returning more points
        than `min_is2_points` (default 100), 28 days, 14 days, etc.
    :type days_r: List[int]
    :param min_is2_points: Minimum number of ICESat-2 points. If there are fewer,
        then the param `days_r` will be increased. If, after the maximum `days_r`
        has been searched and there are still insufficient points, then the
        function will return the available results with a warning. Defaults to 50.
    :type min_is2_points: int
    :param filter_points: If True, then only ICESat-2 points that cover valid
        pixels in the `target_rxd` and `stable_mask` will be returned.
        Defaults to True.
    :type filter_points: bool
    :param include_dt_column: If True, then the datetime column `dt` will be included
        in the returned datafram that shows the temporal offset between the ICESat-2
        point and the provided `date`. Defaults to True.
    :type include_dt_column: bool
    :param epsg: EPSG code for the target_rxd dataset. If `target_rxd` is a tuple,
        this must be provided.
    :type epsg: int
    :param sr_srt: sliderule surface type parameter. 0-land, 1-ocean, 2-seaice,
        3-landice, 4-inlandwater. Defaults to 3.
    :type sr_srt: int
    :param sr_cnf: sliderule confidence threshold.  Default 1 (within 10 m).
    :type sr_cnf: int
    :param sr_ats: sliderule minimum along track spread. Defaults to 20.0 m.
    :type sr_ats: float
    :param sr_cnt: sliderule minimum photon count in segment. Defaults to 10.
    :type sr_cnt: int
    :param sr_len: sliderule length of each extent in metres. Defaults to 20.0.
    :type sr_len: float
    :param sr_res: sliderule step distance for extents. Defaults to 10.0.
    :type sr_res: float
    :param sr_track: sliderule track. 0: all tracks, 1: gt1, 2: gt2, 3: gt3
        Defaults to 0.
    :type sr_track: int
    :param sr_sigma_r_max: sliderule maximum robust dispersion in metres.
        Defaults to 5.
    :type sr_sigma_r_max: int

    :returns: geopandas dataframe of ICESat-2 data
    :rtype: geopandas.DataFrame
    """

    # Sanity check date
    dem_date = to_datetime(date)
    cutoff_date = to_datetime("2018-10-04")
    if dem_date < cutoff_date:
        warn(
            f"You have searched for data beginning {dem_date}. No ICESat-2 data "
            "exists prior to 2018-10-04. This function will return no usable data."
        )

    # Get bounds in right format
    if isinstance(target_rxd, (list, tuple)) and len(target_rxd) == 4:
        if epsg is None:
            raise ValueError(
                "If querying with a tuple of bounds, `epsg` must be defined."
            )
        bounds = target_rxd

    elif isinstance(target_rxd, DataArray):
        bounds = target_rxd.rio.bounds()
        epsg = target_rxd.rio.crs
        # resolution = get_resolution(target_rxd)
    else:
        raise ValueError("`target_rxd` must be xarray object or tuple")

    if isinstance(days_r, int):
        days_r = [days_r]

    # Get search region as shapely geometry in epsg:4326
    gdf_4326 = gpd.GeoDataFrame(geometry=[box(*bounds)], crs=epsg).to_crs(
        4326
    )  # .geometry.values[0]

    # connect to sliderule
    icesat2.init("slideruleearth.io")
    sr_region = sliderule.toregion(gdf_4326)

    # set search parameters (apart from dates)
    params = {
        "poly": sr_region["poly"],
        "srt": sr_srt,  # Surface. 0-land, 1-ocean, 2-seaice, 3-landice (default), 4-inlandwater
        "cnf": sr_cnf,  # Confidence. Default 1 (within 10 m). 2: Low. 3: Medium. 4: High.
        "ats": sr_ats,  # Mininum along track spread. SR Default: 20. pDEMtoold default: 10
        "cnt": sr_cnt,  # Minimum photon count in segment. Default 10.
        "len": sr_len,  # Extent length. ATL06 default is 40 metres.
        "res": sr_res,  # Step distance. ATL06 default is 20 metres.
        "track": sr_track,  # Integer: 0: all tracks, 1: gt1, 2: gt2, 3: gt3. Default 0.
        "sigma_r_max": sr_sigma_r_max,  # Max robust dispersion [m]. Default 5.
    }

    first_iteration = True

    for r in days_r:

        if first_iteration:
            first_iteration = False
        else:
            print(f"Expanding temporal search.")

        date_min = to_datetime(date) - datetime.timedelta(days=r)
        date_max = to_datetime(date) + datetime.timedelta(days=r)
        date_min = date_min.strftime("%Y-%m-%d")
        date_max = date_max.strftime("%Y-%m-%d")

        params["t0"] = date_min
        params["t1"] = date_max

        print(f"Querying points within {r} days... ", end="")

        gdf = icesat2.atl06p(params).to_crs(epsg)

        n_points = len(gdf)

        # Determine number of points that intersect with valid data (according
        # to NaN values and mask)
        if n_points > 0:
            # point coordinates dataset
            coords_ds = Dataset(
                {
                    "x": (["points"], gdf.geometry.x.values),
                    "y": (["points"], gdf.geometry.y.values),
                }
            )

            gdf_flt = gdf.copy()

            # Determine raster value and mask at each point
            gdf_flt["values"] = target_rxd.interp(coords_ds, method="nearest").values
            if stable_mask is not None:
                gdf_flt["mask"] = stable_mask.interp(coords_ds, method="nearest").values

            # Filter values at each point
            gdf_flt = gdf_flt[~gdf_flt["values"].isnull()]
            if stable_mask is not None:
                gdf_flt = gdf_flt[gdf_flt["mask"] > 0]

            # Determine remaining number of points
            n_points_flt = len(gdf_flt)
        else:
            gdf_flt = gdf
            n_points_flt = n_points

        # Print statement
        if n_points == 0:
            print(f"No points found intersecting.", end=" ")
        elif n_points_flt < min_is2_points:
            print(
                f"{n_points} points found with {n_points_flt} intersecting dataset. Below minimum threshold ({min_is2_points}).",
                end=" ",
            )
        else:
            print(f"{n_points} points found with {n_points_flt} intersecting dataset.")
            break

    # Return filtered points if requested
    if n_points > 0:

        if filter_points:
            gdf = gdf_flt

            # Drop filter column values
            gdf.drop(["values"], axis=1, inplace=True)
            if stable_mask is not None:
                gdf.drop(["mask"], axis=1, inplace=True)

        if include_dt_column is True:
            # include 'dt' column
            gdf["request_date_dt"] = gdf.index - to_datetime(date)

    if n_points < min_is2_points:
        warn("Returning GeoDataFrame with number of points below minimum threshold.")
        return gdf
    else:
        return gdf



"""Custom 'accessor' for xarray objects, to allow for filtering, geomorphological
functions, and simple coregistration.
"""

from warnings import warn
from typing import Optional, Literal

import datetime

import xarray as xr
import numpy as np

from xarray import DataArray, Dataset
from geopandas import GeoDataFrame
from pandas import to_datetime
from rasterio.enums import Resampling
from cv2 import connectedComponentsWithStats, CV_32S

from ._geomorphometry import (
    p,
    q,
    r,
    s,
    t,
    slope,
    aspect,
    hillshade,
    plan_curvature,
    horizontal_curvature,
    vertical_curvature,
    mean_curvature,
    gaussian_curvature,
    unsphericity_curvature,
    minimal_curvature,
    maximal_curvature,
)
from ._coreg import coregister, shift_dem  # coregisterdems, coregisterpoints
from ._utils import get_resolution, geospatial_match


LIST_METHODS = ["Florinsky", "ZevenbergThorne"]
LIST_ATTRIBUTES = [
    "slope",
    "aspect",
    "hillshade",
    "plan_curvature",
    "horizontal_curvature",
    "vertical_curvature",
    "horizontal_excess",
    "vertical_excess",
    "mean_curvature",
    "gaussian_curvature",
    "unsphericity_curvature",
    "minimal_curvature",
    "maximal_curvature",
]
LIST_REQUIRING_SECOND_ORDER = [
    "plan_curvature",
    "horizontal_curvature",
    "vertical_curvature",
    "horizontal_excess",
    "vertical_excess",
    "mean_curvature",
    "gaussian_curvature",
    "unsphericity_curvature",
    "minimal_curvature",
    "maximal_curvature",
]


@xr.register_dataarray_accessor("pdt")
class DemAccessor:
    """This class provides ``rioaxarray``-compatible ``DataArray`` objects with access
    to custom functions and methods.

    Functions are accessible via the ``pdt`` accessor: for instance, if you have a DEM
    loaded as the variable ``dem``, you are able to access the ``terrain`` function via
    ``dem.pdt.terrain()``.
    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._center = None

    def geoid_correct(self, geoid: DataArray) -> DataArray:
        """Geoid correct the DEM, given a geoid as an xarray DataArray. The function
        will attempt to reproject the geoid to match if it does not already to do.

        :param geoid: A suitable geoid, as a (rio)xarray dataarray

        :returns: geoid-corrected (rio)xarray dataarray
        :rtype: DataArray"""

        if not geospatial_match(self._obj, geoid):
            geoid = geoid.rio.reproject_match(self._obj, Resampling.bilinear)

        return self._obj - geoid

    def coregister_is2(
        self,
        points_df: GeoDataFrame,
        stable_mask: Optional[DataArray] = None,
        return_stats: Optional[bool] = False,
        max_horiz_offset: float = 50,
        rmse_step_thresh: float = -0.001,
        max_iterations: int = 5,
        verbose: Optional[bool] = True,
        resample: Optional[bool] = True,
        resample_res: Optional[float] = 10,
    ) -> DataArray:
        """Coregisters the scene against reference ICESat-2 data based on the Nuth and Kääb (2011)
        method, as implemented by the PGC.

        Original code available here:
        https://github.com/PolarGeospatialCenter/setsm_postprocessing_python/blob/master/lib/scenes2strips.py

        :points_df: GeoDataFrame of ICESat-2 points to use for coregistration, as provided
            by the `data.load.icesat2_atl06()` function.
        :type points_df: GeoDataFrame
        :param stable_mask: Mask dataarray where 1 = stable region to be used for
            coregistration. If none, coregisters based on the entire scene.
        :type stable_mask: xarray.DataArray
        :param return_stats: If true, returns a metadata dictionary alongside
            the coregistered DEM dataarray. The dictionary contains the coregistration
            status (coregistered, dz_only, failed), the transformation parameter,
            the translation error, and the RMS error. Defaults to False
        :type return_stats: bool
        :param max_horiz_offset: maximum horizontal offset, beyond which XY
            coregistration is ignored and only Z values are corrected. Defaults
            to 15.
        :type max_horiz_offset: float
        :param rmse_step_thresh: threshold for the change in RMSE between two consecutive
            iterations. Defaults to -0.001.
        :type rmse_step_thresh: float
        :param max_iterations: maximum number of iterations. Defaults to 5.
        :type max_iterations: int
        :param verbose: If True, print status updates. Defaults to True.
        :type verbose: bool
        :param resample: If True, resample the DEM to a larger resolution. Defaults to True.
        :type resample: bool
        :param resample_res: Resolution to resample the DEM to. Defaults to 10.
        :type resample_res: float
        """

        # 0. SANITY CHECK
        if stable_mask is None:
            print(
                "No `stable_mask` provided: assuming entire scene is suitable reference surface."
            )
        else:
            check_match = geospatial_match(self._obj, stable_mask, return_info=True)
            if check_match is not True:
                raise ValueError(
                    "Input DEM and stable_mask do not share geospatial information: "
                    f"{check_match}. Consider padding (`.rio.pad_box`) or reprojecting "
                    "(`.rio.repoject_match`) stable_mask."
                )

        # 1. PRE-PROCESSING: INPUT DEM

        resolution = get_resolution(self._obj)

        if (resample is True) and (resolution < resample_res):

            print(
                f"Upsampling {resolution} m DEM to {resample_res} m resolution"
                " for coregistration process..."
            )

            dem_in = self._obj.squeeze().rio.reproject(
                self._obj.rio.crs,  # Maintain the original coordinate reference system
                resolution=resample_res,  # Set the target resolution to 10 meters
                resampling=Resampling.average,  # Upscale with average resampling method
            )

            dem_in_vals = dem_in.values

            if stable_mask is not None:

                mask_in = stable_mask.squeeze().rio.reproject(
                    self._obj.rio.crs,
                    resolution=resample_res,
                    resampling=Resampling.nearest,
                )

                mask_in_vals = mask_in.values

            else:
                mask_in_vals = None

            x_in_vals = dem_in.x.values
            y_in_vals = dem_in.y.values
            resolution = resample_res
            resampled = True

        else:

            dem_in = self._obj.squeeze()
            dem_in_vals = dem_in.values

            if stable_mask is not None:
                mask_in_vals = stable_mask.squeeze().values
            else:
                mask_in_vals = None

            x_in_vals = dem_in.x.values
            y_in_vals = dem_in.y.values
            resampled = False

        # 2. PRE-PROCESSING: REFERENCE POINTS
        # Construct points array [x, y, z]. Can include all the points -
        # coregister function will filter out NaN and masked values.
        points_vals = [
            points_df.geometry.x.values,
            points_df.geometry.y.values,
            points_df.h_mean.values,
        ]

        # 3. COREGISTRATION
        new_dem_array, metadata_dict = coregister(
            dem_in_vals,
            points_vals,
            x_in_vals,
            y_in_vals,
            resolution,
            mask_in_vals,
            max_horiz_offset=max_horiz_offset,
            rmse_step_thresh=rmse_step_thresh,
            max_iterations=max_iterations,
            verbose=verbose,
        )

        # ADD ICESAT-2 SPECIFIC TO METADATA
        if "request_date_dt" in points_df.columns:

            if metadata_dict["coreg_status"] == "failed":
                metadata_dict["points_dt_days_max"] = None
                metadata_dict["points_dt_days_count"] = None

            else:
                dt_days_max = int(
                    points_df["request_date_dt"].dt.round("d").dt.days.abs().max()
                )
                metadata_dict["points_dt_days_max"] = dt_days_max

                dt_days = points_df["request_date_dt"].dt.round("d").dt.days
                dt_days_counts_dict = dt_days.value_counts().to_dict()
                metadata_dict["points_dt_days_count"] = dt_days_counts_dict

        metadata_dict["coregistration_type"] = "reference_icesat2"

        # APPLY COREGISTRATION TO NON-RESAMPLED DATASET
        if resampled:

            # If the coregistration estimation has been run on a DEM that has been
            # upsampled to 10 m, we have to apply the calculated offsets to original
            # high-resolution image

            if metadata_dict["coreg_status"] == "failed":
                new_dem_array = self._obj.squeeze().values.astype("float32")

            else:
                pn = [
                    metadata_dict["z_offset"],
                    metadata_dict["x_offset"],
                    metadata_dict["y_offset"],
                ]

                # Apply offsets
                if pn[1] != 0 and pn[2] != 0:
                    print("Applying transform to original image...")
                    new_dem_array = shift_dem(
                        self._obj.squeeze().values,
                        pn,
                        self._obj.squeeze().x.values,
                        self._obj.squeeze().y.values,
                    ).astype("float32")

                else:
                    print("Applying dz to original image...")
                    new_dem_array = (self._obj.squeeze() - pn[0]).astype("float32")

        # Return coregistered DEM
        if return_stats is True:
            return (self._obj.fillna(0) * 0 + new_dem_array, metadata_dict)

        else:
            return self._obj.fillna(0) * 0 + new_dem_array

    def coregister(
        self,
        reference: DataArray,
        stable_mask: Optional[DataArray] = None,
        return_stats: Optional[bool] = False,
        max_horiz_offset: float = 50,
        rmse_step_thresh: float = -0.001,
        max_iterations: int = 5,
        verbose: Optional[bool] = True,
    ) -> DataArray:
        """This function is deprecated, and redirects to the `coregister_dems`
        function instead. Please consult the `coregister_dems` function for
        documentation.
        """

        warn(
            "The `coregister` function is deprecated and will be removed in a "
            "future version. Please use coregister_dems instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self.coregister_dems(
            reference,
            stable_mask,
            return_stats,
            max_horiz_offset,
            rmse_step_thresh,
            max_iterations,
            verbose,
        )

    def coregister_dems(
        self,
        reference: DataArray,
        stable_mask: Optional[DataArray] = None,
        return_stats: Optional[bool] = False,
        max_horiz_offset: float = 50,
        rmse_step_thresh: float = -0.001,
        max_iterations: int = 5,
        verbose: Optional[bool] = True,
    ) -> DataArray:
        """
        Coregisters the scene against a reference DEM based on the Nuth and Kääb (2011)
        method, as implemented by the PGC.

        Original code available here:
        https://github.com/PolarGeospatialCenter/setsm_postprocessing_python/blob/master/lib/scenes2strips.py

        :param reference: reference DEM DataArray of same extent and resolution.
        :type reference: xarray.DataArray
        :param stable_mask: Mask dataarray where 1 = stable region to be used for
            coregistration. If none, coregisters based on the entire scene.
        :type stable_mask: xarray.DataArray
        :param return_stats: If true, returns a metadata dictionary alongside
            the coregistered DEM dataarray. The dictionary contains the coregistration
            status (coregistered, dz_only, failed), the transformation parameter,
            the translation error, and the RMS error. Defaults to False
        :type return_stats: bool
        :param max_horiz_offset: maximum horizontal offset, beyond which XY
            coregistration is ignored and only Z values are corrected. Defaults
            to 15.
        :type max_horiz_offset: float
        :param rmse_step_thresh: break coregistration loop if rmse step is above
            threshold, defaults to -0.001.
        :type rmse_step_thresh: float
        :param max_iterations: max iterations to attempt, defaults to 5.
        :type max_iterations: int
        :param verbose: Print detailed progress statements, defaults to True.
        :type verbose: bool

        :returns: coregistered (rio)xarray dataarray
        :rtype: DataArray
        :returns: metadata dictionary (if return_stats=True)
        :rtype: dict
        """

        check_match = geospatial_match(self._obj, reference, return_info=True)
        if not check_match is True:
            raise ValueError(
                "Input DEM and reference DEM do not share geospatial information: "
                f"{check_match}. Consider padding (`.rio.pad_box`) or reprojecting "
                "(`.rio.repoject_match`) DEMs."
            )

        if stable_mask is None:
            print(
                "No `stable_mask` provided: assuming entire scene is suitable "
                "reference surface."
            )
            stable_mask = self._obj * 0 + 1
        else:
            check_match = geospatial_match(self._obj, stable_mask, return_info=True)
            if not check_match is True:
                raise ValueError(
                    "Input DEM and stable_mask do not share geospatial information: "
                    "{check_match}. Consider padding (`.rio.pad_box`) or reprojecting "
                    "(`.rio.repoject_match`) stable_mask."
                )

        resolution = get_resolution(self._obj)

        new_dem_array, metadata_dict = coregister(
            self._obj.values,
            reference.values,
            reference.x.values,
            reference.y.values,
            resolution,
            stable_mask.values,
            max_horiz_offset=max_horiz_offset,
            rmse_step_thresh=rmse_step_thresh,
            max_iterations=max_iterations,
            verbose=verbose,
        )

        metadata_dict["coregistration_type"] = "reference_dem"

        # print(f"Translating: {trans[0]:.2f} X, {trans[1]:.2f} Y, {trans[2]:.2f} Z")
        # new_dem_array = shift_dem(self._obj.values, trans)

        if return_stats is True:
            return (self._obj.fillna(0) * 0 + new_dem_array, metadata_dict)

        else:
            return self._obj.fillna(0) * 0 + new_dem_array

    def terrain(
        self,
        attribute: Optional[str | list[str]] = "all",
        method: Literal["Florinsky", "ZevenbergThorne"] = "Florinsky",
        degrees: Optional[bool] = True,
        resolution: Optional[float] = None,
        hillshade_multidirectional: Optional[bool] = False,
        hillshade_altitude: Optional[float] = 45.0,
        hillshade_azimuth: Optional[float] = 315.0,
        hillshade_z_factor: Optional[float] = 2.0,
    ) -> DataArray | Dataset:
        """
        Returns terrain attributes as an xarray DataSet. Available attributes are as
        follows:

         - Slope (`slope`)
         - Aspect (`aspect`)
         - Hillshade (`hillshade`)
         - Plan curvature (`plan_curvature`)
         - Horizontal curvature (`horizontal_curvature`)
         - Vertical curvature (`vertical_curvature`)
         - Horizontal excess curvature (`horizontal_excess`)
         - Vertical excess curvature (`vertical_excess`)
         - Mean curvature (`mean_curvature`)
         - Gaussian curvature (`gaussian_curvature`)
         - Unsphericity curvature (`unsphericity_curvature`)
         - Minimal curvature (`minimal_curvature`)
         - Maximal curvature (`maximal_curvature`)

        Explanations of the above attributes, and their potential applications, can
        be found in Florinsky (2017; doi.org/10.1177/0309133317733667)

        By default, computation of partial derivatives for geomorphometric variable
        calculation is performed following Florinsky (2009), who calculates the third-
        order partial derivative from a 5 x 5 window, which may be more appropriate
        for high-resolution DEMs with some amount of noise. The `method` parameter
        allows for the selection of a more traditional method using a second-order
        derivative from a 3 x 3 window (`ZevenbergThorne`), as outlined by Zevenbergen
        and Throrne (1987).

        :param attribute: The attribute(s) to calculate, as a string or list.
        :type attribute: str | list
        :param method: Method to calculate geomorphometric parameters: "Florisnky" or
            "ZevenbergThorne", defaults to "Florinsky".
        :type method: Literal["Florinsky", "ZevenbergThorne"]
        :param resolution: The DEM resolution. Calculated if empty, defaults to None
        :type resolution: float
        :param degrees: Outputs degrees if True, radians if False, defaults to True
        :type degrees: bool
        :param hillshade_multidirectional: Calculates multidirectional hillshade
            following USGS/GDAL method (Mark, 1992). Defaults to False.
        :type hillshade_multidirectional: bool
        :param hillshade_altitude: Hillshade altitude in degrees (0-90°) from vertical,
            defaults to 45˚
        :type hillshade_altitude: float
        :param hillshade_azimuth: Hillshade azimuth in degrees (0-360°) clockwise from
            North, defaults to 315
        :type hillshade_azimuth: float
        :param hillshade_z_factor: Vertical exaggeration factor, defaults to 1
        :type hillshade_z_factor: float

        :returns: selected terrain attributes as a (rio)xarray DataSet
        :rtype: DataSet
        """

        # Validate attribute(s), resolution, and method
        if isinstance(attribute, str):
            if attribute.lower() == "all":
                attribute = LIST_ATTRIBUTES
            else:
                attribute = [attribute]

        for a in attribute:
            if a.lower() not in LIST_ATTRIBUTES:
                raise ValueError(
                    f"`{a}` is not a valid attribute. Atributes be from list {LIST_ATTRIBUTES}"
                )

        if resolution is None:
            resolution = get_resolution(self._obj)

        if method.lower() not in [m.lower() for m in LIST_METHODS]:
            raise ValueError(
                f"Geomorphometric method must be one of: {LIST_METHODS}. Currently {method}."
            )

        if not isinstance(hillshade_multidirectional, bool):
            raise ValueError("`hillshade_multidirectional` must be boolean")

        # calculate p and q
        p_arr = p(self._obj.values, resolution, method).astype(np.float32)
        q_arr = q(self._obj.values, resolution, method).astype(np.float32)

        # if requested variables require second order variables, calculate them
        attrs_requiring_second_order = [
            attr for attr in attribute if attr in LIST_REQUIRING_SECOND_ORDER
        ]
        if len(attrs_requiring_second_order) >= 1:
            r_arr = r(self._obj.values, resolution, method).astype(np.float32)
            s_arr = s(self._obj.values, resolution, method).astype(np.float32)
            t_arr = t(self._obj.values, resolution, method).astype(np.float32)

        # Calculate attributes
        if ("slope" in attribute) or (
            ("hillshade" in attribute) & (hillshade_z_factor == 1)
        ):
            slope_arr = slope(p_arr, q_arr)

        if (
            ("aspect" in attribute)
            or ("hillshade" in attribute)
            or (hillshade_multidirectional is True)
        ):
            aspect_arr = aspect(p_arr, q_arr)

        if "hillshade" in attribute:
            if hillshade_multidirectional is False:
                # calculate unidirectional hillshade
                if hillshade_z_factor == 1:
                    hillshade_arr = hillshade(
                        np.rad2deg(slope_arr),
                        np.rad2deg(aspect_arr),
                        hillshade_altitude,
                        hillshade_azimuth,
                        norm=True,
                    )
                if hillshade_z_factor != 1:
                    hillshade_arr = hillshade(
                        np.rad2deg(
                            slope(
                                p(
                                    self._obj.values * hillshade_z_factor,
                                    resolution,
                                    method,
                                ),
                                q(
                                    self._obj.values * hillshade_z_factor,
                                    resolution,
                                    method,
                                ),
                            )
                        ),
                        np.rad2deg(aspect_arr),
                        hillshade_altitude,
                        hillshade_azimuth,
                        norm=True,
                    )
            else:
                # calculate multidirectional hillshade following
                # https://pubs.usgs.gov/of/1992/of92-422/of92-422.pdf

                # calculate weights
                w_225 = np.sin(aspect_arr - np.deg2rad(225)) ** 2
                w_270 = np.sin(aspect_arr - np.deg2rad(270)) ** 2
                w_315 = np.sin(aspect_arr - np.deg2rad(315)) ** 2
                w_360 = np.sin(aspect_arr - np.deg2rad(360)) ** 2

                if hillshade_z_factor == 1:
                    slope_input = slope_arr

                if hillshade_z_factor != 1:
                    slope_input = slope(
                        p(
                            self._obj.values * hillshade_z_factor,
                            resolution,
                            method,
                        ),
                        q(
                            self._obj.values * hillshade_z_factor,
                            resolution,
                            method,
                        ),
                    )

                hs = (
                    w_225
                    * hillshade(
                        np.rad2deg(slope_input),
                        np.rad2deg(aspect_arr),
                        60,
                        225,
                        norm=False,
                    )
                    + w_270
                    * hillshade(
                        np.rad2deg(slope_input),
                        np.rad2deg(aspect_arr),
                        60,
                        270,
                        norm=False,
                    )
                    + w_315
                    * hillshade(
                        np.rad2deg(slope_input),
                        np.rad2deg(aspect_arr),
                        60,
                        315,
                        norm=False,
                    )
                    + w_360
                    * hillshade(
                        np.rad2deg(slope_input),
                        np.rad2deg(aspect_arr),
                        60,
                        360,
                        norm=False,
                    )
                ) / 2

                del w_225, w_270, w_315, w_360

                # get normalised hillshade with darkest points as 0
                # hillshade_arr = 1 - (hs - np.nanmin(hs)) / (
                hillshade_arr = (hs - np.nanmin(hs)) / (np.nanmax(hs) - np.nanmin(hs))
                del hs

        if "plan_curvature" in attribute:
            plan_curvature_arr = plan_curvature(p_arr, q_arr, t_arr, r_arr, s_arr)

        if ("horizontal_curvature" in attribute) or ("horizontal_excess" in attribute):
            horizontal_curvature_arr = horizontal_curvature(
                p_arr, q_arr, t_arr, r_arr, s_arr
            )

        if ("vertical_curvature" in attribute) or ("vertical_excess" in attribute):
            vertical_curvature_arr = vertical_curvature(
                p_arr, q_arr, t_arr, r_arr, s_arr
            )

        if (
            ("mean_curvature" in attribute)
            or ("unsphericity_curvature" in attribute)
            or ("maximal_curvature" in attribute)
            or ("minimal_curvature" in attribute)
            or ("horizontal_excess" in attribute)
            or ("vertical_excess" in attribute)
        ):
            mean_curvature_arr = mean_curvature(p_arr, q_arr, t_arr, r_arr, s_arr)

        if (
            ("gaussian_curvature" in attribute)
            or ("unsphericity_curvature" in attribute)
            or ("maximal_curvature" in attribute)
            or ("minimal_curvature" in attribute)
            or ("horizontal_excess" in attribute)
            or ("vertical_excess" in attribute)
        ):
            gaussian_curvature_arr = gaussian_curvature(
                p_arr, q_arr, t_arr, r_arr, s_arr
            )

        if (
            ("unsphericity_curvature" in attribute)
            or ("maximal_curvature" in attribute)
            or ("minimal_curvature" in attribute)
            or ("horizontal_excess" in attribute)
            or ("vertical_excess" in attribute)
        ):
            unsphericity_curvature_arr = unsphericity_curvature(
                mean_curvature_arr, gaussian_curvature_arr
            )

        if (
            ("minimal_curvature" in attribute)
            or ("horizontal_excess" in attribute)
            or ("vertical_excess" in attribute)
        ):
            minimal_curvature_arr = minimal_curvature(
                mean_curvature_arr, unsphericity_curvature_arr
            )

        if "maximal_curvature" in attribute:
            maximal_curvature_arr = maximal_curvature(
                mean_curvature_arr, unsphericity_curvature_arr
            )

        # Then convert necessary values into degrees from radians...
        if degrees is True:
            if "slope" in attribute:
                slope_arr = np.rad2deg(slope_arr)
            if "aspect" in attribute:
                aspect_arr = np.rad2deg(aspect_arr)

        # Then convert into xarray dataset:
        xds = self._obj.to_dataset(name="dem", promote_attrs=True)

        if "slope" in attribute:
            xds["slope"] = self._obj * 0 + slope_arr
        if "aspect" in attribute:
            xds["aspect"] = self._obj * 0 + aspect_arr
        if "hillshade" in attribute:
            xds["hillshade"] = self._obj * 0 + hillshade_arr

        if "plan_curvature" in attribute:
            xds["plan_curvature"] = self._obj * 0 + plan_curvature_arr
        if "horizontal_curvature" in attribute:
            xds["horizontal_curvature"] = self._obj * 0 + horizontal_curvature_arr
        if "vertical_curvature" in attribute:
            xds["vertical_curvature"] = self._obj * 0 + vertical_curvature_arr
        if "horizontal_excess" in attribute:
            xds["horizontal_excess"] = self._obj * 0 + (
                horizontal_curvature_arr - minimal_curvature_arr
            )
        if "vertical_excess" in attribute:
            xds["vertical_excess"] = self._obj * 0 + (
                vertical_curvature_arr - minimal_curvature_arr
            )
        if "mean_curvature" in attribute:
            xds["mean_curvature"] = self._obj * 0 + mean_curvature_arr
        if "gaussian_curvature" in attribute:
            xds["gaussian_curvature"] = self._obj * 0 + gaussian_curvature_arr
        if "unsphericity_curvature" in attribute:
            xds["unsphericity_curvature"] = self._obj * 0 + unsphericity_curvature_arr
        if "minimal_curvature" in attribute:
            xds["minimal_curvature"] = self._obj * 0 + minimal_curvature_arr
        if "maximal_curvature" in attribute:
            xds["maximal_curvature"] = self._obj * 0 + maximal_curvature_arr

        # If you have only asked for one attribute, return a dataarray rather than a
        # dataset
        if len(attribute) == 1:
            xds = xds[attribute[0]]

        return xds

    def mask_ocean(
        self,
        candidate_height_thresh_m: Optional[float] = 10,
        candidate_area_thresh_km2: Optional[float] = 1,
        near_sealevel_thresh_m: Optional[float] = 5,
        return_sealevel_as_zero: Optional[bool] = False,
        return_mask: Optional[bool] = False,
    ) -> DataArray:
        """Returns a DEM with mélange/ocean regions filtered out, adapting the method
        of Shiggins _et al._ (2023). If no likely sea level is identified, returns the
        original DEM.

        WARNING: The DEM must be geoid-corrected for this function to work correctly.

        :param candidate_height_thresh_m: Maximum value relative to geoid to be considered
            as SL, in m, defaults to 10
        :type candidate_height_thresh_m: float
        :param candidate_area_thresh_km2: Minimum area beneath `candidate_height_thresh_m`
            to be considered for sea level assessment, in km^2. Defaults to 1
        :type candidate_area_thresh_km2: float
        :param near_sealevel_thresh_m: Filter out regions below this value, in metres above
            sea level. Defaults to 5
        :type near_sealevel_thresh_m: float
        :param return_sealevel_as_zero: If True, subtracts that estimated sea level from
            the DEM, returning new height values with the estimated sea level set to 0
            m. Necessary for calculating accurate iceberg heights. Defaults to False.
        :type return_sealevel_as_zero: bool
        :param return_mask: Return the sea level mask rather than the masked DEM.
            Defaults to False.
        :type return_mask: bool

        :returns: Filtered DEM as xarray DataArray
        :rtype: DataArray
        """

        # resolution = get_resolution(self._obj)

        # mask = self._obj.pdt.get_ocean_mask(
        #     candidate_height_thresh_m,
        #     candidate_area_thresh_km2,
        #     near_sealevel_thresh_m,
        # )

        est_sea_level = self._obj.pdt.get_sea_level(
            candidate_height_thresh_m,
            candidate_area_thresh_km2,
        )

        if est_sea_level is None:
            mask = None
        else:
            mask = self._obj > (est_sea_level + near_sealevel_thresh_m)

        if return_mask is True:
            if mask is None:
                warn("No sea level detected in image. Returning `None` as mask object.")
            return mask
        else:
            if mask is None:
                warn("No sea level detected in image. Returning original DEM")
                return self._obj
            else:
                if return_sealevel_as_zero is True:
                    return self._obj.where(mask) - est_sea_level
                elif return_sealevel_as_zero is False:
                    return self._obj.where(mask)
                else:
                    raise ValueError("return_sealevel_as_zero must be True or False")

    def get_sea_level(
        self,
        candidate_height_thresh_m: Optional[float] = 10,
        candidate_area_thresh_km2: Optional[float] = 1,
    ) -> float:
        """Returns estimated sea level following method of Shiggins _et al._ (2023). If no
        candidate sea level is identified, None is returned. DEM must be geoid-corrected.

        :param candidate_height_thresh_m: Maximum value relative to geoid to be considered
            as SL, in m, defaults to 10
        :type candidate_height_thresh_m: float
        :param candidate_area_thresh_km2: Minimum area beneath `candidate_height_thresh_m`
            to be considered for sea level assessment, in km^2, defaults to 1
        :type candidate_area_thresh_km2: float

        :returns: Estimated sea level
        :rtype: float
        """

        resolution = get_resolution(self._obj)

        # Get values close to sea level as 1D numpy array
        near_sealevel_values = self._obj.values.ravel()
        near_sealevel_values = near_sealevel_values[
            near_sealevel_values < candidate_height_thresh_m
        ]

        # Skip melange filtering if the candidate region is less than the threshold area
        thresh_px_n = int(candidate_area_thresh_km2 * 1e6 / (resolution * 2))
        if len(near_sealevel_values) < thresh_px_n:
            return None

        # Else, construct a 0.25m resolution histogram between -15 and +15:
        else:
            bin_edges = np.arange(-15.125, 15.375, 0.25)
            bin_centres = bin_edges[:-1] + 0.125
            hist, _ = np.histogram(near_sealevel_values, bins=bin_edges)

            # The estimated sea level is the centre of the modal bin
            est_sea_level = bin_centres[np.argmax(hist)]

            return est_sea_level

    def mask_icebergs(
        self,
        area_thresh_m2: Optional[float | int] = 1e6,
        retain_icebergs: Optional[bool] = False,
        return_mask: Optional[bool] = False,
        connectivity: Literal[4, 8] = 4,
    ) -> DataArray:
        """After masking the ocean using the `mask_ocean()` function, icebergs will
        remain. This function gives you the opportunity to mask them as well by
        identifying the size of connected groups of pixels and masking above/below a
        threshold.

        By default, this function masks icebergs and retains terrestrial ice/land.
        However, by setting `retain_icebergs = True`, the function will instead mask
        ice/land and retain icebergs, allowing iceberg area and volume to be assessed.
        For accurate volume assessment, the DEM 0 m value must be set to the estimated
        sea level height from the `get_sea_level()` function.  This can be automated by
        setting `return_sealevel_as_zero = True` in the mask_ocean() function.

        :param area_thresh_m2: Size threshold between icebergs and terrestrial ice/land, in
            m2. Defaults to 1e6 m2 (1 km2)
        :type area_thresh_m2: float | int
        :param retain_icerbergs: By default, this function masks icebergs and retains
            terrestrial ice/land (by masking pixel groups below the threshold area). By
            switching the parameter to True, the function will instead mask out larger
            pixel groups and retain the smaller icebergs.
        :type retain_icebergs: bool
        :param return_mask: If True, returns mask of icebergs (where icebergs = 1). If
            False, returns input DEM with icebergs masked. Defaults to False.
        :type return_mask: bool
        :type area_thresh_m2: float | int
        :param connectivity: Connectivity with which to calculate iceberg regions. Must
            be 4 or 8. Defaults to 4.
        :type connectivity: int

        :returns: Either masked DEM or mask as xarray DataArray, depending on
            the input value of `return_mask`.
        :rtype: DataArray

        """

        resolution = get_resolution(self._obj)

        # Generate boolean mask in int8 format for cv2.connectedComponentsWithStats()
        connected_input = (~np.isnan(self._obj.values)).astype(np.int8).squeeze()

        # Generate labels and statistics
        num_labels, labels, stats, _ = connectedComponentsWithStats(
            connected_input, connectivity, CV_32S
        )

        # calculate areas of connected components from number of pixels and resolution
        areas = stats[:, 4] * resolution * resolution

        # Get labels that are greater than area_thesh
        if (retain_icebergs is True) and (return_mask is False):
            labels_valid = np.where(
                areas < area_thresh_m2, np.arange(num_labels), np.nan
            )
        else:
            labels_valid = np.where(
                areas > area_thresh_m2, np.arange(num_labels), np.nan
            )

        labels_valid = labels_valid[~np.isnan(labels_valid)]

        if return_mask is True:
            return self._obj * 0 + ~np.isin(labels, labels_valid)

        else:
            return self._obj.where(np.isin(labels, labels_valid))



"""This module contains functions necessary to open an ArcticDEM/REMA strip as an
xarray DataArray, from either local or AWS sources.
"""

import os
from typing import Optional, Literal, Union
from importlib import resources

import rioxarray as rxr
import geopandas as gpd
import numpy as np

from rioxarray.merge import merge_arrays
from xarray import DataArray
from shapely.geometry.polygon import Polygon
from geopandas.geodataframe import GeoDataFrame

from ._utils import clip

# arctic and rema valid versions for STAC retrival
VERSIONS = {"arcticdem": ["v3.0", "v4.1"], "rema": ["v2.0"]}

# filenames of mosaic indexes in ./src/pdemtools/mosaic_index directory
ARCTICDEM_V3_INDEX_FNAME = "ArcticDEM_Mosaic_Index_v3_gpkg.gpkg"
ARCTICDEM_V3_INDEX_2M_LAYER_NAME = "ArcticDEM_Mosaic_Index_v3_2m"
ARCTICDEM_V4_INDEX_FNAME = "ArcticDEM_Mosaic_Index_v4_1_gpkg.gpkg"
ARCTICDEM_V4_INDEX_2M_LAYER_NAME = "ArcticDEM_Mosaic_Index_v4_1_2m"
REMA_V2_INDEX_FNAME = "REMA_Mosaic_Index_v2_gpkg.gpkg"
REMA_V2_INDEX_2M_LAYER_NAME = "REMA_Mosaic_Index_v2_2m"

# aws location
PREFIX = "https://pgc-opendata-dems.s3.us-west-2.amazonaws.com"

# valid mosaic resolutions
VALID_MOSAIC_RES = ["2m", "10m", "32m"]


def from_fpath(
    dem_fpath: str,
    bounds: Optional[Union[tuple, Polygon]] = None,
    bitmask_fpath: Optional[str] = None,
    pad: Optional[bool] = False,
    chunks: Optional[Union[int, tuple, dict]] = None,
) -> DataArray:
    """Given a filepath (local or an AWS link), loads the desired ArcticDEM/REMA DEM
    strip as an ``xarray`` ``DataArray``. Option to filter to bounds and bitmask, if
    provided.

    If AWS link is provided, strip will be downloaded from the relevant AWS bucket. 2 m
    DEM strips are large in size and loading remotely from AWS may take some time.

    :param dem_fpath: Filepath of DEM strip
    :type dem_fpath: str
    :param bounds: Clip to bounds [xmin, ymin, xmax, ymax], in EPSG:3413 (ArcticDEM) or
        EPSG:3031 (REMA). Will accept a shapely geometry to extract bounds from.
        Defaults to None
    :type bounds: tuple | Polygon, optional
    :param mask_fpath: Path to _bitmask.tif file used to mask the DEM, defaults to None
    :type mask_fpath: str, optional
    :param pad: If the DEM strip is not the full extent of the given bounds,
        pad with NaNs to match the full bounds. Defaults to False.
    :type pad: bool
    :param chunks: Chunk size for `rioxarray.open_rasterio`, triggering `dask`
        parallelisation and lazy evaluation/loading. Defaults to `None`.
    :type chunks: int | tuple | dict

    :returns: xarray DataArray of DEM strip
    :rtype: DataArray
    """

    # Sanitise input
    if not isinstance(pad, bool):
        raise ValueError("pad must be True or False")

    # Open dataarray using rioxarray
    dem = rxr.open_rasterio(dem_fpath, chunks=chunks)

    # Convert shapely geometry to bounds
    if isinstance(bounds, Polygon):
        bounds = bounds.bounds

    # Clip if requested, or get whole bounds if not
    if bounds is not None:
        dem = clip(dem, bounds)
    else:
        bounds = dem.rio.bounds()
        if pad is not False:
            raise ValueError("If `bounds` is not provided, `pad` must be False")

    # Filter -9999.0 values
    dem = dem.where(dem > -9999.0)

    # Mask using bitmask if requested
    if bitmask_fpath is not None:
        mask = rxr.open_rasterio(bitmask_fpath)
        if bounds is not None:
            mask = clip(mask, bounds)
        dem = dem.where(mask == 0)
        del mask

    # Remove `band` dim
    dem = dem.squeeze(drop=True)

    # Enforce CF-compliant names
    dem["x"].attrs["axis"] = "X"
    dem["x"].attrs["long_name"] = "x coordinate of projection"
    dem["x"].attrs["standard_name"] = "projection_x_coordinate"
    dem["x"].attrs["units"] = "metre"

    dem["y"].attrs["axis"] = "Y"
    dem["y"].attrs["long_name"] = "y coordinate of projection"
    dem["y"].attrs["standard_name"] = "projection_y_coordinate"
    dem["y"].attrs["units"] = "metre"

    if pad is True:
        dem = dem.rio.pad_box(*bounds, constant_values=np.nan)

    return dem


def preview(
    row: GeoDataFrame,
    bounds: Optional[Union[tuple, Polygon]] = None,
    bitmask: Optional[bool] = True,
    pad: Optional[bool] = False,
):
    """Loads a 10 m hillshade preview of the desired ArcticDEM/REMA DEM strip as an
    ``xarray`` ``DataArray``, for preliminary plotting and assessment. Option to filter
    to bounds.

    :param row: A selected row from the GeoDataFrame output of pdemtools.search. Can
        either select a row manually using gdf.isel[[i]] where `i` is the desired row,
        or provide the entire GeoDataFrame,
    :type row: GeoDataFrame
    :param bounds: Clip to bounds [xmin, ymin, xmax, ymax], in EPSG:3413 (ArcticDEM) or
        EPSG:3031 (REMA). Will accept a shapely geometry to extract bounds from.
        Defaults to None
    :type bounds: tuple | Polygon, optional
    :param bitmask: If True, loads the masked hillshade according to the PGC bitmask.
        If False, loads the unmasked hillshade. Defaults to True
    :type bitmask: bool
    :param pad: If the DEM strip is not the full extent of the given bounds,
        pad with NaNs to match the full bounds. Defaults to True.
    :type pad: bool

    :returns: xarray DataArray of DEM strip 10 m hillshade
    :rtype: DataArray
    """

    if bitmask is True:
        try:
            hillshade_10m_url = row.href_hillshade_masked.values[0]
        except:
            hillshade_10m_url = row.href_hillshade_masked
    elif bitmask is False:
        try:
            hillshade_10m_url = row.href_hillshade.values[0]
        except:
            hillshade_10m_url = row.href_hillshade
    else:
        raise ValueError("masked must be True or False")

    preview = from_fpath(hillshade_10m_url, bounds)

    if bounds is not None and pad is True:
        preview = preview.rio.pad_box(*bounds, constant_values=np.nan)

    return preview.where(preview > 0)


def from_search(
    row: GeoDataFrame,
    bounds: Optional[Union[tuple, Polygon]] = None,
    bitmask: Optional[bool] = True,
    pad: Optional[bool] = False,
    chunks: Optional[Union[int, tuple, dict]] = None,
):
    """Given a row from the GeoDataFrame output of ``pdemtools.search()``, loads the 2
    m DEM strip of the desired ArcticDEM/REMA DEM strip as an xarray DataArray.

    Downloads from the relevant AWS bucket, as an xarray ``DataArray``. 2 m DEM strips
    are large in size and loading remotely from AWS may take some time. Option to
    filter to bounds and bitmask.

    :param row: A selected row from the GeoDataFrame output of pdemtools.search. Select
        the row manually using ``row = gdf.isel[[i]]``, or loop through the rows using
        ``for _, row in gdf.iterrows():``.
    :type row: GeoDataFrame
    :param bounds: Clip to bounds [xmin, ymin, xmax, ymax], in EPSG:3413 (ArcticDEM) or
        EPSG:3031 (REMA). Will accept a ``shapely.Polygon`` geometry to extract bounds
        from. Defaults to None
    :type bounds: tuple | Polygon, optional
    :param bitmask: Choose whether apply the associated bitmask. Defaults to True
    :type bitmask: bool, optional
    :param pad: If the DEM strip is not the full extent of the given bounds,
        pad with NaNs to match the full bounds. Defaults to False.
    :type pad: bool
    :param chunks: Chunk size for `rioxarray.open_rasterio`, triggering `dask`
        parallelisation and lazy evaluation/loading. Defaults to `None`.
    :type chunks: int | tuple | dict

    :returns: xarray DataArray of DEM strip
    :rtype: DataArray
    """

    try:
        dem_url = row.href_dem.values[0]
    except:
        dem_url = row.href_dem

    # try:
    #     s3url = row.s3url.values[0]
    # except:
    #     s3url = row.s3url

    # json_url = "https://" + s3url.split("/external/")[1]
    # dem_url = json_url.replace(".json", "_dem.tif")

    # Construct bitmask fpath, if required
    if bitmask is True:
        try:
            bitmask_url = row.href_mask.values[0]
        except:
            bitmask_url = row.href_mask
    else:
        bitmask_url = None

    # Pass AWS URL locations to load_local command
    return from_fpath(
        dem_url,
        bounds,
        bitmask_url,
        pad,
        chunks,
    )


def from_id(
    dataset: Literal["arcticdem", "rema"],
    geocell: str,
    dem_id: str,
    bounds: Optional[Union[tuple, Polygon]] = None,
    bitmask: Optional[bool] = True,
    bucket: Optional[str] = "https://pgc-opendata-dems.s3.us-west-2.amazonaws.com",
    version: Optional[str] = "s2s041",
    preview: Optional[bool] = False,
    pad: Optional[bool] = False,
    chunks: Optional[Union[int, tuple, dict]] = None,
) -> DataArray:
    """An alternative method of loading the selected ArcticDEM/REMA strip, which
    requires only the geocell and the dem_id (e.g. ``geocell = 'n70w051'``, ``dem_id =
    'SETSM_s2s041_WV01_20200709_102001009A689B00_102001009B63B200_2m_lsf_seg2'``).

    Downloads from the relevant AWS bucket, as an xarray ``DataArray``. Option to
    filter to bounds and bitmask. 2 m DEM strips are large in size and loading
    remotely from AWS may take some time.

    NOTE: This function is kept for vestigial purposes, and may be removed or
    significantly altered in a future release.

    :param dataset: Either 'arcticdem' or 'rema'. Case-insensitive.
    :type dataset: str
    :param geocell: Geographic grouping of ArcticDEM / REMA strip. e.g. 'n70w051'.
    :type geocell: str
    :param dem_id: ArcticDEM/REMA strip ID. e.g.
        'SETSM_s2s041_WV01_20200709_102001009A689B00_102001009B63B200_2m_lsf_seg2'
    :type dem_id: str
    :param bounds: Clip to bounds [xmin, ymin, xmax, ymax], in EPSG:3413 (ArcticDEM) or
        EPSG:3031 (REMA), defaults to None
    :type bounds: tuple, optional
    :param bitmask: Choose whether apply the associated bitmask, defaults to True
    :type bitmask: bool, optional
    :param bucket: AWS buck link, defaults to
        'https://pgc-opendata-dems.s3.us-west-2.amazonaws.com'
    :type bucket: str
    :param version: Version string, defaults to 's2s041'
    :type version: str
    :param preview: Return just a link to the STAC preview page, defaults to False
    :type preview: bool, optional
    :param pad: If the DEM strip is not the full extent of the given bounds,
        pad with NaNs to match the full bounds. Defaults to False.
    :type pad: bool
    :param chunks: Chunk size for `rioxarray.open_rasterio`, triggering `dask`
        parallelisation and lazy evaluation/loading. Defaults to `None`.
    :type chunks: int | tuple | dict

    :return: xarray DataArray of DEM strip
    :rtype: DataArray
    """

    # Sanitise data
    dataset = dataset.lower()
    geocell = geocell.lower()

    if preview is True:
        browser_prefix = "https://polargeospatialcenter.github.io/stac-browser/#/external/pgc-opendata-dems.s3.us-west-2.amazonaws.com"
        preview_fpath = (
            f"{browser_prefix}/{dataset}/strips/{version}/2m/{geocell}/{dem_id}.json"
        )
        return preview_fpath

    # Construct DEM fpath
    dem_fpath = f'{bucket}/{dataset}/"strips"/{version}/2m/{geocell}/{dem_id}_dem.tif'
    # dem_fpath = os.path.join(
    #     bucket, dataset, "strips", version, "2m", geocell, f"{dem_id}_dem.tif"
    # )

    # Construct bitmask fpath, if required
    if bitmask is True:
        bitmask_fpath = (
            f'{bucket}/{dataset}/"strips"/{version}/2m/{geocell}/{dem_id}_bitmask.tif'
        )
        # bitmask_fpath = os.path.join(
        #     bucket, dataset, "strips", version, "2m", geocell, f"{dem_id}_bitmask.tif"
        # )
    else:
        bitmask_fpath = None

    # Pass AWS URL locations to load_local command
    return from_fpath(
        dem_fpath,
        bounds,
        bitmask_fpath,
        pad,
        chunks,
    )


def mosaic(
    dataset: Literal["arcticdem", "rema"],
    resolution: Literal["2m", "10m", "32m"],
    bounds: Union[tuple, Polygon] = None,
    version: Optional[Literal["v2.0", "v3.0", "v4.1"]] = None,
    chunks: Optional[Union[int, tuple, dict]] = None,
):
    """Given a dataset, resolution, and bounding box, downloads the ArcticDEM or REMA
    mosiac from AWS.

    :param dataset: The desired dataset, either 'arcticdem' or 'rema'.
        Case-instensitive.
    :type datasat: str
    :param resolution: The desired mosaic resolution to download - must be either ``2m``,
        ``10m``, or ``32m`` (will also accept ``2``, ``10``, and ``32`` as ``int`` types)
    :type resolutions: str | int
    :param version: Desired ArcticDEM or REMA version. Must be a valid version available
        from the PGC STAC API (e.g. ``v3.0`` or ``v4.1`` for ArcticDEM, or ``v2.0`` for REMA).
    :type version: str
    :param bounds: Clip to bounds [xmin, ymin, xmax, ymax], in EPSG:3413 (ArcticDEM) or
        EPSG:3031 (REMA). Will accept a shapely geometry to extract bounds from.
    :type bounds: tuple | Polygon, optional
    :param chunks: Chunk size for `rioxarray.open_rasterio`, triggering `dask`
        parallelisation and lazy evaluation/loading. Defaults to `None`.
    :type chunks: int | tuple | dict

    :return: xarray DataArray of DEM mosaic
    :rtype: DataArray
    """

    # sanity check that datset and versioning is correct versioning is valid for selected dataset
    dataset = dataset.lower()
    if dataset not in VERSIONS.keys():
        raise ValueError(
            f"Dataset must be one of {VERSIONS.keys}. Currently `{dataset}`."
        )
    if version is None:  # pick the most recent dataset
        version = VERSIONS[dataset][-1]
    else:
        if version not in VERSIONS[dataset]:
            raise ValueError(
                f"Version of {dataset} must be one of {VERSIONS[dataset]}. Currently `{version}`"
            )

    # get resolution as str
    if isinstance(resolution, int):
        resolution = f"{resolution}m"

    # check if valid resolution
    if resolution not in VALID_MOSAIC_RES:
        raise ValueError(
            f"Resolution must be one of {VALID_MOSAIC_RES}. Currently `{resolution}`"
        )

    # Sanitise shapely geometry to bounds tuple
    if isinstance(bounds, Polygon):
        bounds = bounds.bounds

    # get dataset version
    if dataset == "arcticdem" and version == "v3.0":
        layer = ARCTICDEM_V3_INDEX_2M_LAYER_NAME
    elif dataset == "arcticdem" and version == "v4.1":
        layer = ARCTICDEM_V4_INDEX_2M_LAYER_NAME
    elif dataset == "rema" and version == "v2.0":
        layer = REMA_V2_INDEX_2M_LAYER_NAME
    else:
        raise ValueError(
            "Cannot retrive internal index filepath for specified dataset and version."
        )

    # Load tiles that intersect with AOI
    index_fpath = _get_index_fpath(dataset, version=version)
    with open(index_fpath, "rb") as file:
        tiles = gpd.read_file(index_fpath, layer=layer, bbox=bounds, engine="fiona")

        if len(tiles) < 1:
            raise ValueError(
                f"No {dataset} mosaic tiles found to intersect with bounds {aoi}"
            )

        # get aws filepaths from the tiles dataframe
        fpaths = []
        for _, row in tiles.iterrows():
            fpath = _aws_link(
                row, dataset=dataset, version=version, resolution=resolution
            )
            fpaths.append(fpath)

        tiles = None  # release tiles object

    # remove duplicates in 10m and 32m (which load supertiles, not tiles)
    fpaths = list(set(fpaths))

    # load dem(s)
    dems = []
    for fpath in fpaths:
        dem = rxr.open_rasterio(fpath, chunks=chunks).rio.clip_box(*bounds)
        dems.append(dem)

    if len(fpaths) == 1:
        dem = rxr.open_rasterio(fpaths[0], chunks=chunks).rio.clip_box(*bounds)

    # If multiple dems, merge them - NB I don't know whether this breaks lazy
    # evaluation for chunked data
    if len(dems) > 1:
        dem = merge_arrays(dems)
    else:
        dem = dems[0]

    dems = None  # release dem objects for memory management

    # Filter -9999.0 values to np.nan
    dem = dem.where(dem > -9999.0)

    # Remove `band` dim
    dem = dem.squeeze(drop=True)

    # Enforce CF-compliant names
    dem["x"].attrs["axis"] = "X"
    dem["x"].attrs["long_name"] = "x coordinate of projection"
    dem["x"].attrs["standard_name"] = "projection_x_coordinate"
    dem["x"].attrs["units"] = "metre"

    dem["y"].attrs["axis"] = "Y"
    dem["y"].attrs["long_name"] = "y coordinate of projection"
    dem["y"].attrs["standard_name"] = "projection_y_coordinate"
    dem["y"].attrs["units"] = "metre"

    return dem


def _get_index_fpath(
    dataset: Literal["arcticdem", "rema"],
    version: Literal["v2.0", "v3.0", "v4.0"],
):
    """Given ``arcticdem`` or ``rema``, gets the filepath of the package dataset using the
    ``importlib`` library. ARCTICDEM and REMA global variables necessary.
    """

    # get dataset version
    if dataset == "arcticdem" and version == "v3.0":
        fname = ARCTICDEM_V3_INDEX_FNAME
    elif dataset == "arcticdem" and version == "v4.1":
        fname = ARCTICDEM_V4_INDEX_FNAME
    elif dataset == "rema" and version == "v2.0":
        fname = REMA_V2_INDEX_FNAME
    else:
        raise ValueError(
            "Cannot retrive internal index filepath for specified dataset and version."
        )

    return resources.files("pdemtools.mosaic_index").joinpath(fname)


def _aws_link(
    row,
    dataset: Literal["arcticdem", "rema"],
    version: str,
    resolution: Literal["2m", "10m", "32m"],
    prefix: Optional[str] = PREFIX,
):
    """Using inputs from mosaic() function and the AWS location from the global variable
    `PREFIX`, construct the filepath of the relevant ArcticDEM or REMA mosaic tile.
    """
    # Construct appropriate suffix, considering ArcticDEM v3.0's alternate naming scheme
    if dataset == "arcticdem" and version == "v3.0":
        suffix = f"_{resolution}_{version}_reg_dem.tif"
    else:
        suffix = f"_{resolution}_{version}_dem.tif"

    # Construct appropriate filename given resolution
    if resolution == "2m":
        fname = f"{row.tile}{suffix}"
    elif resolution in ["10m", "32m"]:
        fname = f"{row.supertile}{suffix}"
    else:
        raise ValueError(f"Input `resolution` must be one of ['2m', '10m', '32m']")

    # Return appropriate filepath.
    return f"{prefix}/{dataset}/mosaics/{version}/{resolution}/{row.supertile}/{fname}"



"""
This module contains simple geospatial utility functions used across the other modules.
"""

import rioxarray as rxr

from xarray import DataArray

def get_resolution(xds: DataArray) -> float:
    """Retrive resolution from rioxarray Dataset. Must be the same in `x` and `y`
    direction.

    :param dem_xds: xarray Dataset of DEM strip
    :type dem_xds: Dataset

    :returns: Resolution of DEM strip
    :rtype: float
    """

    resolutions = xds.rio.resolution()
    if abs(resolutions[0]) != abs(resolutions[1]):
        raise ValueError("Dataset resolution not the same in `x` and `y` directions")
    else:
        return abs(resolutions[0])


def geospatial_match(
    rxd_1: DataArray, rxd_2: DataArray, return_info: bool = False
) -> bool:
    """Check whether two (rio)xarray DataArrays or Datasets have the same geospatial
    information.

    :param rxd_1: First (rio)xarray Dataset to compare
    :type rxd_1: DataArray
    :param rxd_2: Second (rio)xarray Dataset to compare
    :type rxd_2: DataArray

    :returns: True if Datasets match, False if not
    :rtype: bool
    """

    failed = []

    if not rxd_1.rio.shape == rxd_2.rio.shape:
        failed.append("shape")
    if not rxd_1.rio.resolution() == rxd_2.rio.resolution():
        failed.append("resolution")
    if not rxd_1.rio.bounds() == rxd_2.rio.bounds():
        failed.append("bounds")
    if not rxd_1.rio.crs == rxd_2.rio.crs:
        failed.append("crs")

    if len(failed) == 0:
        return True
    else:
        if return_info == True:
            return failed
        else:
            return False


def clip(
    rxd: DataArray,
    bounds: tuple,
) -> DataArray:
    """Clips (rio)xarray DataArray to bounds in format [xmin, ymin, xmax, ymax] using
    the rioxarray clip_box() function. Provide bounds in matching CRS (e.g. EPSG:3413
    for ArcticDEM strips and EPSG:3031 for REMA strips).

    :param rxd: (rio)xarray object to clip
    :type rxd: DataArray
    :param bounds: List of AOI bounds in format [xmin, ymin, xmax, ymax]
    :type bounds: tuple

    :returns: Clipped xarray Dataset
    :rtype: Dataset
    """

    # TODO: More informative error handling of intersection issues than rioxarray
    # currently provides

    return rxd.rio.clip_box(*bounds)



import pytest

import rioxarray as rxr
import xarray as xr
import geopandas as gpd
import numpy as np

import pdemtools as pdt

from numpy.testing import assert_allclose
from shapely.geometry import box


def generate_test_data_1():
    """Generates a 2D sin pattern"""

    x = np.arange(-500, 500, 2)
    y = np.arange(-500, 500, 2)

    X, Y = np.meshgrid(x, y)
    Z = np.sin((X**2 + Y**2) / np.max(X**2) * 10) * 10 + 30

    dem = xr.DataArray(
        data=Z,
        coords=[("x", np.flip(x)), ("y", y)],
        dims=["y", "x"],  # Adjust order if "x" and "y" represent column coordinates
    )
    dem.rio.write_crs("EPSG:3413", inplace=True)
    dem.rio.write_transform(inplace=True)  # Set the georeferencing

    return dem


def generate_test_data_2(x_offset=-13.4, y_offset=+14.6, z_offset=15.3):
    """As for `generate_test_data_1` but offsets by fixed amount"""

    x = np.arange(-500, 500, 2)
    y = np.arange(-500, 500, 2)

    x2 = np.arange(-500 + x_offset, 500 + x_offset, 2)
    y2 = np.arange(-500 + y_offset, 500 + y_offset, 2)

    X, Y = np.meshgrid(x2, y2)
    Z = np.sin((X**2 + Y**2) / np.max(X**2) * 10) * 10 + 30 + z_offset

    dem = xr.DataArray(
        data=Z,
        coords=[("x", np.flip(x)), ("y", y)],
        dims=["y", "x"],  # Adjust order if "x" and "y" represent column coordinates
    )
    dem.rio.write_crs("EPSG:3413", inplace=True)
    dem.rio.write_transform(inplace=True)  # Set the georeferencing

    return dem


def generate_test_data_3(z_offset=-30):
    """Generates a 2D sin pattern"""

    x = np.arange(-10000, 10000, 32)
    y = np.arange(-10000, 10000, 32)

    X, Y = np.meshgrid(x, y)
    Z = np.sin((X**2 + Y**2) / np.max(X**2) * 10) * 10 + 30 + z_offset

    dem = xr.DataArray(
        data=Z,
        coords=[("x", np.flip(x)), ("y", y)],
        dims=["y", "x"],  # Adjust order if "x" and "y" represent column coordinates
    )
    dem.rio.write_crs("EPSG:3413", inplace=True)
    dem.rio.write_transform(inplace=True)  # Set the georeferencing

    return dem


def test_geomorphometry():

    variable_min_max = {
        "slope": (0.0, 24.721811294555664),
        "aspect": (0.0, 360.0),  # (1.3660377589985728e-05, 359.9999694824219),
        "hillshade": (0.0, 1.0),
        "horizontal_curvature": (-0.0007999989320524037, 0.0007856095326133072),
        "vertical_curvature": (-0.027519449591636658, 0.022581202909350395),
        "mean_curvature": (-0.013776100240647793, 0.011318518780171871),
        "gaussian_curvature": (-1.0310398465662729e-05, 9.49732475419296e-06),
        "unsphericity_curvature": (0.0, 0.013750489801168442),
        "minimal_curvature": (-0.02751944772899151, 0.0007855556323193014),
        "maximal_curvature": (-0.0007999998051673174, 0.022581204771995544),
    }

    dem = generate_test_data_1()
    dem = dem.pdt.terrain(variable_min_max.keys())

    for variables in variable_min_max.keys():
        min_val = dem[variables].min().item()
        max_val = dem[variables].max().item()
        print(f"{variables} | min: {min_val} | max: {max_val}")
        assert_allclose(min_val, variable_min_max[variables][0])
        assert_allclose(max_val, variable_min_max[variables][1])


def test_coregistration():

    dem_1 = generate_test_data_1()
    dem_2 = generate_test_data_2()

    mask_gdf = gpd.GeoDataFrame(geometry=[box(-200, -100, 0, 100)], crs=3413)
    mask = pdt.data.bedrock_mask_from_vector(mask_gdf, dem_1)

    dem_2_coreg = dem_2.pdt.coregister_dems(dem_1, mask)

    # diff = (dem_2 - dem_1).mean().item()
    diff_coreg = (dem_2_coreg - dem_1).mean().item()

    # expected_diff = 15.476494902013904
    expected_diff_coreg = 0.19603341170854155

    # print(diff, corrected_diff)

    assert_allclose(expected_diff_coreg, diff_coreg)


def test_geoid_correction():

    dem = generate_test_data_1()
    geoid = dem * 0 + 30

    dem_geoidcor = dem.pdt.geoid_correct(geoid)

    # expected_mean = 30.835231084697448
    expected_geoidcor_mean = 0.8352310846974458

    # mean = dem.mean().item()
    geoidcor_mean = dem_geoidcor.mean().item()

    assert_allclose(expected_geoidcor_mean, geoidcor_mean)


def test_mask_ocean():

    dem = generate_test_data_3(z_offset=-25)
    dem_masked = dem.pdt.mask_ocean(near_sealevel_thresh_m=5)

    orig_min = dem.min().item()
    new_min = dem_masked.min().item()
    expected_min = 0.0014336554957807834

    # print(dem_masked)
    # print(dem_masked.rio.resolution())
    # print(dem.pdt.get_sea_level())
    # print(orig_min)
    # print(new_min)

    assert new_min > orig_min
    assert_allclose(new_min, expected_min)


if __name__ == "__main__":
    test_geomorphometry()
    test_coregistration()
    test_geoid_correction()
    test_mask_ocean()



"""
Functions for calculating geomorphometry variables.
"""

import numba
import numpy as np


"""
Numba functions for quadratic 3x3 polynomial fit following Zevenbergen and Thorne (1987)
"""


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _p_zt(z, w):
    return (-z[0, -1] + z[0, 1]) / (2 * w)


@numba.njit
def p_zt(z, w=2):
    return _p_zt(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _q_zt(z, w):
    return (z[-1, 0] - z[1, 0]) / (2 * w)


@numba.njit
def q_zt(z, w=2):
    return _q_zt(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _r_zt(z, w):
    return ((z[0, -1] + z[0, 1]) / 2 - z[0, 0]) / (w**2)


@numba.njit
def r_zt(z, w=2):
    return _r_zt(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _t_zt(z, w):
    return ((z[-1, 0] + z[1, 0]) / 2 - z[0, 0]) / (w**2)


@numba.njit
def t_zt(z, w=2):
    return _t_zt(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _s_zt(z, w):
    return (1) * (-z[-1, -1] + z[-1, 1] + z[1, -1] - z[1, 1]) / (4 * w**2)


@numba.njit
def s_zt(z, w=2):
    return _s_zt(z, w).astype('float32')


"""
 Numba functions for third order polynomial 5x5 fit following Florinsky (2009)
"""


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _p_f(z, w):
    return (1 / (420 * w)) * (
        44 * (z[-2, 1] + z[2, 1] - z[-2, -1] - z[2, -1])
        + 31
        * (
            z[-2, -2]
            + z[2, -2]
            - z[-2, 2]
            - z[2, 2]
            + 2 * (z[-1, 1] + z[1, 1] - z[-1, -1] - z[1, -1])
        )
        + 17 * (z[0, 2] - z[0, -2] + 2 * (z[0, 1] - z[0, -1]))
        + 5 * (z[-1, 2] + z[1, 2] - z[-1, -2] - z[1, -2])
    )


@numba.njit
def p_f(z, w=2):
    return _p_f(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _q_f(z, w):
    return (1 / (420 * w)) * (
        44 * (z[-1, -2] + z[-1, 2] - z[1, -2] - z[1, 2])
        + 31
        * (
            z[2, -2]
            + z[2, 2]
            - z[-2, -2]
            - z[-2, 2]
            + 2 * (z[-1, -1] + z[-1, 1] - z[1, -1] - z[1, 1])
        )
        + 17 * (z[-2, 0] - z[2, 0] + 2 * (z[-1, 0] - z[1, 0]))
        + 5 * (z[-2, -1] + z[-2, 1] - z[2, -1] - z[2, 1])
    )


@numba.njit
def q_f(z, w=2):
    return _q_f(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _r_f(z, w):
    return (1 / (35 * w**2)) * (
        2
        * (
            z[-2, -2]
            + z[-2, 2]
            + z[-1, -2]
            + z[-1, 2]
            + z[0, -2]
            + z[0, 2]
            + z[1, -2]
            + z[1, 2]
            + z[2, -2]
            + z[2, 2]
        )
        - 2 * (z[-2, 0] + z[-1, 0] + z[0, 0] + z[1, 0] + z[2, 0])
        - z[-2, -1]
        - z[-2, 1]
        - z[-1, -1]
        - z[-1, 1]
        - z[0, -1]
        - z[0, 1]
        - z[1, -1]
        - z[1, 1]
        - z[2, -1]
        - z[2, 1]
    )


@numba.njit
def r_f(z, w=2):
    return _r_f(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _t_f(z, w):
    return (1 / (35 * w**2)) * (
        2
        * (
            z[-2, -2]
            + z[-2, -1]
            + z[-2, 0]
            + z[-2, 1]
            + z[-2, 2]
            + z[2, -2]
            + z[2, -1]
            + z[2, 0]
            + z[2, 1]
            + z[2, 2]
        )
        - 2 * (z[0, -2] + z[0, -1] + z[0, 0] + z[0, 1] + z[0, 2])
        - z[-1, -2]
        - z[-1, -1]
        - z[-1, 0]
        - z[-1, 1]
        - z[-1, 2]
        - z[1, -2]
        - z[1, -1]
        - z[1, 0]
        - z[1, 1]
        - z[1, 2]
    )


@numba.njit
def t_f(z, w=2):
    return _t_f(z, w).astype('float32')


@numba.stencil(func_or_mode="constant", cval=np.nan)
def _s_f(z, w):
    return (1 / (100 * w**2)) * (
        z[-1, 1]
        + z[1, -1]
        - z[-1, -1]
        - z[1, 1]
        + 4 * (z[-2, 2] + z[2, -2] - z[-2, -2] - z[2, 2])
        + 2
        * (
            z[-2, 1]
            + z[-1, 2]
            + z[1, -2]
            + z[2, -1]
            - z[-2, -1]
            - z[-1, -2]
            - z[1, 2]
            - z[2, 1]
        )
    )


@numba.njit
def s_f(z, w=2):
    return _s_f(z, w).astype('float32')


"""
Wrapper functions to choose between Florinsky and Zevenberg methods
"""


def p(z, w, method):
    if method.lower() == "florinsky":
        return p_f(z.squeeze(), w)
    elif method.lower() == "zevenbergthorne":
        return p_zt(z.squeeze(), w)
    else:
        raise ValueError(
            "`method` variable must be one of ['Florinsky', 'ZevenbergThorne']"
        )


def q(z, w, method):
    if method.lower() == "florinsky":
        return q_f(z.squeeze(), w)
    elif method.lower() == "zevenbergthorne":
        return q_zt(z.squeeze(), w)
    else:
        raise ValueError(
            "`method` variable must be one of ['Florinsky', 'ZevenbergThorne']"
        )


def r(z, w, method):
    if method.lower() == "florinsky":
        return r_f(z.squeeze(), w)
    elif method.lower() == "zevenbergthorne":
        return r_zt(z.squeeze(), w)
    else:
        raise ValueError(
            "`method` variable must be one of ['Florinsky', 'ZevenbergThorne']"
        )


def s(z, w, method):
    if method.lower() == "florinsky":
        return s_f(z.squeeze(), w)
    elif method.lower() == "zevenbergthorne":
        return s_zt(z.squeeze(), w)
    else:
        raise ValueError(
            "`method` variable must be one of ['Florinsky', 'ZevenbergThorne']"
        )


def t(z, w, method):
    if method.lower() == "florinsky":
        return t_f(z.squeeze(), w)
    elif method.lower() == "zevenbergthorne":
        return t_zt(z.squeeze(), w)
    else:
        raise ValueError(
            "`method` variable must be one of ['Florinsky', 'ZevenbergThorne']"
        )


"""
Functions for processing curvature from derivatives
"""


def slope(p_arr, q_arr):
    """outputs in radians"""
    return np.arctan(np.sqrt(p_arr**2 + q_arr**2))


def aspect(p_arr, q_arr):
    """outputs in radians"""

    return np.arctan2(p_arr, q_arr) + np.pi

    # return np.deg2rad(
    #     -90 * (1 - np.sign(q_arr)) * (1 - np.abs(np.sign(p_arr)))
    #     + 180 * (1 + np.sign(p_arr))
    #     - (180 / np.pi)
    #     * np.sign(p_arr)
    #     * np.arccos(-q_arr / np.sqrt(p_arr**2 + q_arr**2))
    # )


def hillshade(slope, aspect, altitude=45, azimuth=315, norm=True):
    """accepts degrees"""

    if slope.max() < 2 * np.pi:
        raise Warning(
            "Maximum slope value is < 2π (slope.max()), ensure input is degrees"
        )
    if aspect.max() < 2 * np.pi:
        raise Warning(
            "Maximum slope value is < 2π (aspect.max()), ensure input is degrees"
        )

    hs = 255.0 * (
        (np.cos(np.deg2rad(altitude)) * np.cos(np.deg2rad(slope)))
        + (
            np.sin(np.deg2rad(altitude))
            * np.sin(np.deg2rad(slope))
            * np.cos(np.deg2rad(azimuth) - np.deg2rad(aspect))
        )
    ).astype("float32")

    # hs = np.cos(np.pi * 0.5 - np.deg2rad(aspect) - np.deg2rad(azimuth)) * np.sin(
    #     np.deg2rad(slope)
    # ) * np.sin(np.pi * 0.5 - np.deg2rad(altitude)) + np.cos(np.deg2rad(slope)) * np.cos(
    #     np.pi * 0.5 - np.deg2rad(altitude)
    # )

    # return normalised values
    if norm == True:
        return (hs - np.nanmin(hs)) / (np.nanmax(hs) - np.nanmin(hs))
    else:
        return hs


def plan_curvature(p_arr, q_arr, t_arr, r_arr, s_arr):
    return -(
        ((q_arr**2 * r_arr) - (2 * p_arr * q_arr * s_arr) + (p_arr**2 * t_arr))
        / np.sqrt((p_arr**2 + q_arr**2) ** 3)
    )


def horizontal_curvature(p_arr, q_arr, t_arr, r_arr, s_arr):
    return -(
        ((q_arr**2 * r_arr) - (2 * p_arr * q_arr * s_arr) + (p_arr**2 * t_arr))
        / ((p_arr**2 + q_arr**2) * np.sqrt(1 + p_arr**2 + q_arr**2))
    )


def vertical_curvature(p_arr, q_arr, t_arr, r_arr, s_arr):
    return -(
        ((p_arr**2 * r_arr) + (2 * p_arr * q_arr * s_arr) + (q_arr**2 * t_arr))
        / ((p_arr**2 + q_arr**2) * np.sqrt((1 + p_arr**2 + q_arr**2) ** 3))
    )


def mean_curvature(p_arr, q_arr, t_arr, r_arr, s_arr):
    return -(
        (1 + q_arr**2) * r_arr - 2 * p_arr * q_arr * s_arr + (1 + p_arr**2) * t_arr
    ) / (2 * np.sqrt(1 + p_arr**2 + q_arr**2) ** 3)


def gaussian_curvature(p_arr, q_arr, t_arr, r_arr, s_arr):
    return (r_arr * t_arr - s_arr**2) / ((1 + p_arr**2 + q_arr**2) ** 2)


def unsphericity_curvature(mean, gaussian):
    return np.sqrt(mean**2 - gaussian)


def maximal_curvature(mean, unsphericity):
    return mean + unsphericity


def minimal_curvature(mean, unsphericity):
    return mean - unsphericity



import pytest

import pdemtools as pdt

from importlib import resources

SAMPLE_INDEX_FNAME = "test_arcticdem_index_kiv_steenstrup.parquet"
SAMPLE_INDEX_FPATH = resources.files("pdemtools.test_data").joinpath(SAMPLE_INDEX_FNAME)


def test_search_stac():

    bounds = (459000, -2539000, 473000, -2528000)

    gdf = pdt.search(
        dataset="arcticdem",
        bounds=bounds,
        dates="20170101/20221231",
        months=[6, 7, 8, 9],
        years=[2017, 2021],
        baseline_max_hours=24,
        sensors=["WV03", "WV02", "WV01"],
        accuracy=2,
        min_aoi_frac=0.7,
    )

    assert len(gdf) == 6


def test_search_index():

    bounds = (459000, -2539000, 473000, -2528000)

    gdf = pdt.search(
        dataset="arcticdem",
        bounds=bounds,
        dates="20170101/20221231",
        months=[6, 7, 8, 9],
        years=[2017, 2021],
        baseline_max_hours=24,
        sensors=["WV03", "WV02", "WV01"],
        min_aoi_frac=0.7,
        index_fpath=SAMPLE_INDEX_FPATH,
    )

    assert len(gdf) == 6


if __name__ == "__main__":
    test_search_stac()
    test_search_index()



from importlib.metadata import version

# import submodules
from . import load
from . import data
from ._index_search import search

# import xarray accessor
from ._accessor import DemAccessor

# import hidden modules for test purposes
from . import _coreg
from . import _geomorphometry
from . import _utils

__version__ = "1.1.0"

__all__ = ["search", "DemAccessor"]
